#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import pickle
import pandas as pd
import os
import logging

from utilities.initialise_log import save_log


class check_general_settings():
    
    """
        This class imports the temporary pickle files saved with the
        information provided by the user through the GUI and performs quality
        control on the provided information.
        
        The class instance error then decides whether SHIRE aborts the run
        or if the next step is initialised.
        
        The information on the results of the quality check is saved in a log
        file which stored either at the location of the training dataset or
        the prediction dataset.
    """
    
    def __init__(self):
        
        path = ['tmp_settings.pkl', 'tmp_map.pkl', 'tmp_train.pkl', 'tmp_pred.pkl']
        self.error = False
        
        save_path = 'check_user_input.log'
            
        if os.path.exists(save_path):
            os.remove(save_path)
        
        self.logger = save_log(save_path)
        self.logger.info("Start checking user input")
        self.open_reference()
        
        for self.path in path:
            self.logger.info('Check: ' + self.path)
            if os.path.exists(self.path):
                self.logger.info(self.path + ': available')
                self.open_file()
                self.check_parameters()
                
                if self.path == path[3]:
                    self.check_coordinates()
                if self.path == path[3] or self.path == path[2]:
                    self.check_path_extension_geosummary()
            else:
                self.logger.info(self.path + ': not available')
            
        self.logger.info('Check completed')
        for handler in self.logger.handlers:
            handler.close()
            self.logger.removeHandler(handler)
                
    def open_file(self, path=None):
        
        if path == None:
            path = self.path
            
        with open(path, 'rb') as handle:
            self.params = pickle.load(handle)
        
        if path == None:
            self.logger.info('Parameters loaded')
            
    def open_reference(self):
        
        self.ref = pd.read_csv('./utilities/properties_user_input.csv')
        self.logger.info('Reference imported')
        
    def compare_type(self, key, types):
        
        type_map = {
                'int': int,
                'float': float,
                'str': str,
                'list': list,
                'dict': dict,
                'bool': bool
            }
        
        types = types.split(',')
        types = [type_map.get(i) for i in types]

        if type(self.params[key]) not in types:
            self.logger.error(key + ': Wrong parameter type provided! Provide: ' + str(types))
            self.error = True   
                
    def compare_extension(self, key, ext):
        
        if len(self.params[key].split('.')) != 2:
            self.logger.error(key + ': Path names should not contain full stops!')
            self.error = True
        
        ext = ext.split(',')
        
        if self.params[key].split('.')[1] not in ext:
            self.logger.error(key + ': Wrong file format provided! Provide: ' + str(ext))
            self.error = True

    def compare_range(self, key, r):

        r = r.split(',')
        
        if r[1] == 'inf':
            if self.params[key] < float(r[0]): 
                self.logger.error(key + ': Value not within range!')
                self.error = True                
        else:
            if self.params[key] < float(r[0]) or self.params[key] > float(r[1]):
                self.logger.error(key + ': Value not within range!')
                self.error = True    
                    
    def check_coordinates(self):
        
        if self.params['south'] >= self.params['north']:
            if self.params['south'] == self.params['north']:
                self.logger.error('Careful! South coordinate indentical to north coordinate!')
            else:
                self.logger.error('Careful! South coordinate north of north coordinate!')
            self.error = True
            
        if self.params['west'] >= self.params['east']:
            if self.params['west'] == self.params['east']:
                self.logger.error('Careful! West coordinate identical to east coordinate!')
                self.error = True
            else:
                if ((self.params['west'] < 0 and self.params['west'] > -10) and (self.params['east'] > 0 and self.params['east'] < 10)) or ((self.params['west'] > 0 and self.params['west'] > 170) and (self.params['east'] < 0 and self.params['east'] < -170)):
                    self.logger.warning('Careful! Please check east and west coordinates!')
                else:
                    self.logger.error('Careful! West coordinate east of east coordinate!')
                    self.error = True
        
    def check_file_exist(self, key, path):
        
        if not os.path.exists(path) and not os.path.isdir(os.path.dirname(path)):
            self.logger.error(key + ': Path or file does not exist!')
            self.error = True
        
    def check_path_extension_geosummary(self):
        
        self.logger.info('Check paths in geospatial data summary')
        summary = pd.read_csv(self.params['geo_path'])
        keys_to_include = pd.read_csv(self.params['feat_path'])
        for key in list(keys_to_include['keys_to_include']):
            idx = list(summary['keys']).index(key)
            
            if summary.at[idx, 'path'].split('.')[1] not in ['nc', 'tif', 'tiff']:
                self.logger.error(key + ': Wrong file format!')
                self.error = True
                
            if not os.path.exists(summary.at[idx, 'path']):
                self.logger.error(key + ': File cannot be found!')
                self.error = True        
    
    def check_parameters(self):
        
        ref_keys = self.ref['key'].tolist()
        for key in list(self.params.keys()):
            idx = ref_keys.index(key)
            self.logger.info('Check ' + key + ' | is path: ' + str(self.ref.at[idx, 'path']) + ' | Range: ' + str(self.ref.at[idx, 'range']) + ' | Extension: ' + str(self.ref.at[idx, 'extension']) + ' | Type: ' + str(self.ref.at[idx, 'type']))
            if self.ref.at[idx, 'path'] == 1:
                #print(self.ref.at[idx, 'path'])
                self.check_file_exist(key, self.params[key])
            if self.ref.at[idx, 'range'] != 'None':
                self.compare_range(key, self.ref.at[idx, 'range'])
            if self.ref.at[idx, 'extension'] != 'None' and self.ref.at[idx, 'path'] == 1:
                self.compare_extension(key, self.ref.at[idx, 'extension'])
            if self.ref.at[idx, 'type'] != 'None':
                self.compare_type(key, self.ref.at[idx, 'type'])

            