#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy as np
import pandas as pd
import netCDF4 as nc
import os
import itertools
import pickle
import logging
import tkinter as tk

from tqdm import tqdm
from LatLon23 import LatLon, Latitude, Longitude
from sklearn.cluster import KMeans
from tkinter import Label, ttk

from utilities.import_format import import_nc, import_tif
from utilities.cut_and_interpolate_gui import cut_and_interpolate
from utilities.strings_for_ncfile import char_to_string
from utilities.handle_categorical_values import handle_categorical_values
from data_preprocessing_gui import generate_data_matrix

class create_training_data:

    """
        This class generates the training dataset for
        the Random Forest Classifier.

        Input:
            from_scratch: boolean, True if training dataset should be
            generated from scratch, otherwise False
            delete: True if dataset/feature should be deleted from csv-file
                    False if dataset should be added to existing csv-file
                    (careful: from_scratch needs to be False!)
            data_to_handle: list of features that should be added/deleted
                            datasets need to be listed in list_of_raw_datasets
            cluster: boolean, True if training locations are to be clustered
            interpolation: boolean, True if datasets are supposed to be
            interpolated before extracting information
            preprocessing: list of length equal to data_to_handle,
            preprocessing methods for the individual datasets,
            can be either 'no_interpolation', 'interpolation' or 'cluster'

        Output:
            csv file
    """

    def __init__(self, master, log=None):

        self.num_clusters = 15
        self.df_train = 'not_set_yet'
        self.logger = log
        self.import_settings()
        
        self.datasets_summary = pd.read_csv(self.properties_train['geo_path'])

        self.row = 0

        self.master = master
        self.master.geometry()
        self.master.winfo_toplevel().title("Training dataset generation")

        Label(self.master, text="Log:").grid(row=self.row, column=0)
        self.row = self.row + 1

        Label(self.master, text="Training dataset generation started").grid(
            row=self.row, column=1)
        self.logger.info('Training dataset generation started')
        self.row = self.row + 1
        self.master.update()

        if self.from_scratch:
            # If all datasets are supposed to be preprocessed
            # using the same method
            #if self.how_to_preprocess is None:
            if self.from_scratch:
                self.logger.info(
                    'Launch training dataset generation according to desired approach')
            else:
                self.logger.info(
                    'Launch feature addition according to desired approach')
            if self.cluster is True and self.interpolation is True:
                Label(self.master, text="Features to include:").grid(
                    row=self.row, column=1)
                self.logger.info("Features to include:")
                self.row = self.row + 1
                Label(self.master, text=self.data_to_handle).grid(
                    row=self.row, column=1)
                self.logger.info(self.data_to_handle)
                self.row = self.row + 1
                self.master.update()
                self.main_cluster()
            elif self.cluster is False and self.interpolation is True:
                self.main()
            elif self.interpolation is False:

                Label(self.master, text="Features to include:").grid(
                    row=self.row, column=1)
                self.logger.info("Features to include:")
                self.row = self.row + 1
                Label(self.master, text=self.data_to_handle).grid(
                    row=self.row, column=1)
                self.logger.info(self.data_to_handle)
                self.row = self.row + 1
                self.master.update()

                self.main_no_interpolation()
    
        elif not self.from_scratch and self.delete:
    
            self.logger.info('Feature deletion started')
    
            self.import_input_training_dataset()
            self.delete_feature()
            self.save_training_data()
    
        elif not self.from_scratch and not self.delete:
    
            self.logger.info('Feature addition started')
    
            self.import_input_training_dataset()
            self.add_feature()
            self.save_training_data()

    def import_settings(self):
        
        """
            User input is imported and the settings of the run are defined
        """

        with open('tmp_train.pkl', 'rb') as handle:
            self.properties_train = pickle.load(handle)
        self.overview = pd.read_csv(self.properties_train['geo_path'])

        if self.properties_train['from_scratch'] == 1:
            self.from_scratch = True
            self.delete = False
        elif self.properties_train['delete'] == 1:
            self.from_scratch = False
            self.delete = True
        if self.properties_train['add'] == 1:
            self.from_scratch = False
            self.delete = False
            
        self.data_to_handle = pd.read_csv(
            self.properties_train['feat_path'])
        self.data_to_handle = list(self.data_to_handle['keys_to_include'])

        if self.properties_train['interpolation'] == 1:
            self.interpolation = True
            self.cluster = False
            self.no_interpolation = False
        elif self.properties_train['cluster'] == 1:
            self.cluster = True
            self.interpolation = False
            self.no_interpolation = False
        elif self.properties_train['no_interpolation'] == 1:
            self.cluster = False
            self.interpolation = False
            self.no_interpolation = True

        with open('tmp_settings.pkl', 'rb') as handle:
            self.properties_settings = pickle.load(handle)
        self.properties_settings['train_path'] = self.properties_train[
            'train_path']
        
        self.logger.info('User input has been imported')


    def delete_feature(self):

        """
            Features in data_to_handle are deleted from the training dataset
        """

        to_drop = []
        for feat in self.data_to_handle:
            for col in self.df_train.columns.tolist():
                if feat in col:
                    to_drop.append(col)
            
        self.df_train.drop(columns=to_drop, inplace=True)

        self.logger.info('Feature deleted')
        self.logger.info(
            'Training dataset now includes the following features:')
        self.logger.info(str(self.df_train.columns))

    def import_cube(self):

        """
            Import data cube created in data_preprocessing.py
        """

        self.ds = nc.Dataset(self.s.outfile)
        self.x = self.ds['Longitude'][:].data
        self.y = self.ds['Latitude'][:].data
        self.cube = self.ds['Result'][:, :, :].data

        features = self.ds['features'][:].data

        self.features = char_to_string(features)

        Label(self.master, text="Cube was imported").grid(
            row=self.row, column=1)
        self.row = self.row + 1
        self.master.update()
        Label(self.master, text="Cube contains the following features:").grid(
            row=self.row, column=1)
        self.row = self.row + 1
        Label(self.master, text=self.features).grid(row=self.row, column=1)
        self.row = self.row + 1
        self.master.update()

        self.logger.info('Cube with feature datasets imported')
        self.logger.info('Cube contains the following features:')
        self.logger.info(self.features)

    def import_landslide_database(self):

        """
            Import training dataset which needs to be provided as csv file
            The landslide instances are supplemented with absence locations.
            Missing values and Nan values are removed.
        """

        if self.properties_train['ls_path'].split('.')[-1] != 'csv':
            self.logger.error(
                'Error: Please provide a csv-file as input dataset')

        self.df_train = pd.read_csv(self.properties_train['ls_path'])

        for column in list(self.df_train.columns):
            if column not in [self.properties_train['id'],
                              self.properties_train['x'],
                              self.properties_train['y']]:
                self.df_train.drop(inplace=True, labels=column, axis=1)
        self.check_ls_inventory()
        self.length_before_cleaning = len(self.df_train)
        self.add_nonls_locations()
        self.df_train = pd.concat([self.df_train, self.df_absence],
                                  axis=0,
                                  ignore_index=True)
        self.df_train = self.df_train.rename(
            columns={self.properties_train['x']: 'xcoord',
                     self.properties_train['y']: 'ycoord',
                     self.properties_train['id']: 'ID'})
        self.df_train['label'] = self.label_training_data()

        Label(self.master,
              text="Historic landslides were imported").grid(
                  row=self.row, column=1)
        self.row = self.row + 1

        self.logger.info('Landslide inventory imported')
        self.logger.info('Currently the following features are included:')
        self.logger.info(str(self.df_train.columns))

    def add_nonls_locations(self):

        """
            Supplement presence data with absence data. It needs to be
            pre-generated (see nonls_events_coordinates.py).
        """

        if self.properties_train['nonls_path'].split('.')[-1] == 'csv':
            self.absence = pd.read_csv(self.properties_train['nonls_path'])
            
            self.absence = self.absence.rename(
            columns={
                self.properties_train['x_nonls']: self.properties_train['x'],
                self.properties_train['y_nonls']: self.properties_train['y']})

            nonls_id = [
                'nonls_event_' + str(i) for i in range(len(self.absence))]
            self.absence.insert(0, self.properties_train['id'], nonls_id)

            self.logger.info('Absence locations added')

        elif self.properties_train['nonls_path'].split('.')[-1] == 'nc':
            ds = nc.Dataset(self.properties_train['nonls_path'])

            x = ds[self.properties_train['x_nonls']][:].data
            y = ds[self.properties_train['y_nonls']][:].data

            self.df_absence = pd.DataFrame(index=range(len(x)),
                                           columns=list(self.df_train.columns))

            self.df_absence[self.properties_train['id']] = [
                'nonls_event_' + str(i) for i in range(len(x))]
            self.df_absence[self.properties_train['x']] = list(x)
            self.df_absence[self.properties_train['y']] = list(y)

            self.logger.info('Absence locations added')

    def label_training_data(self):

        """
            Provide labels to the training data
        """

        label = [int(1) for i in range(len(self.df_train))]
        label[(len(label)-len(self.df_absence)):len(label)] = [
            int(0) for i in range(len(self.df_absence))]

        self.logger.info('Training data labeled')

        return label

    def import_input_training_dataset(self):

        """
            Existing training dataset is imported.
        """

        # Check if training dataset exists and import it
        if not os.path.exists(self.properties_train['train_path']):
            self.logger.error('Training dataset does not exist yet.\
                              Please generate from scratch.')
        else:
            self.df_train = pd.read_csv(self.properties_train['train_path'])
            self.features = list(self.df_train.columns)
            self.logger.info('Training dataset imported')
            self.logger.info('Features included in the training dataset:')
            self.logger.info(self.features)

    def add_feature(self):

        """
            Add feature to the training dataset
        """

        self.logger.info('Feature adding started')

        # Features contained in the training dataset
        cols = list(self.df_train.columns)
        # List of labels for basic information
        basic = ['Ereignis-Nr', 'xcoord', 'ycoord', 'label']
        print(self.data_to_handle)
        if not self.cluster and self.interpolation:
            # Call data_preprocessing.py
            self.s = generate_data_matrix(
                from_scratch=False,
                delete=False,
                dataset='training',
                data_to_handle=self.data_to_handle,
                geo_overview=self.overview,
                settings=self.properties_settings,
                settings_train_pred=self.properties_train,
                keys_already_included=[x for x in cols if x not in basic])

            # Import generated cube of cut and interpolated datasets
            self.import_cube()

            for count, key in enumerate(self.data_to_handle):
                # Delete feature if it already exists in training dataset
                if key in self.df_train.columns:
                    self.logger.warning('Feature already exists in Dataset.\
                                        Existing feature is deleted')
                    self.df_train = self.df_train.drop(key, axis=1)

                self.logger.info('Adding ' + key + '...')
                # Create empty DataFrame
                if count == 0:
                    self.df_features = pd.DataFrame(
                        index=range(len(self.df_train)),
                        columns=self.data_to_handle)

                for index, row in self.df_train.iterrows():
                    x_ind = int(
                        (np.abs(
                            self.x
                            - row[self.properties_train['x']])).argmin())
                    y_ind = int(
                        (np.abs(
                            self.y
                            - row[self.properties_train['y']])).argmin())

                    self.df_features.at[index, key] = self.cube[
                        y_ind,
                        x_ind,
                        self.features.index(key)]

            # Combine old training dataset with additional features
            self.df_train = pd.concat([self.df_train, self.df_features],
                                      axis=1)
            # Adapt column order
            self.logger.info('Feature successfully added')

            self.clean_df()
            self.handle_categorical_features(var=self.data_to_handle)
            self.logger.info('One-hot encoding completed')
            
        elif self.cluster:
            for key in self.data_to_handle:
                if key in self.df_train.columns:
                    self.logger.info(
                        'Feature already exists in dataset.\
                            Existing feature is deleted')
                    self.df_train = self.df_train.drop(key, axis=1)

            self.main_cluster()
            self.logger.info('Feature successfully added')
        elif not self.cluster and not self.interpolation:
            self.main_no_interpolation()
            self.logger.info('Feature successfully added')

        self.logger.info('Training dataset contains following features:')
        self.logger.info(str(self.df_train.columns))

    def extract_gridded_info(self):

        """
            Extraction of the information of the geospatial datasets at all
            elements of the training dataset.

            If training data is located within prediction area and if area is
            small enough no further interpolation is necessary.
        """

        self.logger.info('Extraction of gridded information started')

        self.df_features = pd.DataFrame(index=range(len(self.df_train)),
                                        columns=self.features)

        # Iterate over all instances of the training dataset
        # and extract geospatial information
        k = 0
        self.open_secondary_window(kind='int', maximum=len(self.df_train))
        for index, row in tqdm(self.df_train.iterrows()):

            self.progress_var.set(k)
            self.label.config(text=str(index)
                              + '/' + str(len(self.df_train))
                              + ' instance')
            self.master.update()
            k += 1

            # Indices of training data elements are determined
            x_ind = int((np.abs(self.x - row['xcoord'])).argmin())
            y_ind = int((np.abs(self.y - row['ycoord'])).argmin())

            tmp = list(self.cube[y_ind, x_ind, :])
            self.df_features.loc[index] = tmp

        self.progress_var.set(k)
        self.label.config(text=str(index)
                          + '/' + str(len(self.df_train))
                          + ' instance')
        self.master.update()

        Label(self.master, text="Gridded information was extracted").grid(
            row=self.row, column=1)
        self.row = self.row + 1
        self.master.update()

        self.logger.info('Extraction of gridded information completed')

    def check_ls_inventory(self):

        """
            Rows are removed with missing values or nan
        """

        rows_with_missing_values = self.df_train[self.df_train.isnull().any(axis=1)].index.tolist()
        
        if len(set(rows_with_missing_values)) > 0:
            self.df_train.drop(index=set(rows_with_missing_values), inplace=True)
            self.logger.info(str(len(set(rows_with_missing_values)))
                             + ' rows are removed due to missing or nan values')

    def clean_df(self):

        """
            Rows are removed from the dataset where no_value given in settings
            occurs.
        """

        self.logger.info('Cleaning of training dataset started')

        Label(self.master, text='Cleaning in progress').grid(
            row=self.row, column=1)
        self.row = self.row + 1
        self.master.update()

        count = 0
        ind = []
        for col in self.df_train:
            ind.append([i for i, x in enumerate(
                list(self.df_train[col])) if x == self.properties_settings[
                    'no_value']])
            count = count + 1

        ind = list(itertools.chain.from_iterable(ind))
        ind = list(set(ind))

        self.logger.info(str(len(ind))
                         + ' rows will be removed due to invalid data')

        Label(self.master, text=str(len(ind))
              + ' rows will be removed due to invalid data').grid(
                  row=self.row, column=1)
        self.row = self.row + 1
        self.master.update()

        self.df_train.drop(index=ind, inplace=True)
        self.df_train.reset_index(drop=True, inplace=True)

        self.logger.info('Cleaning of training dataset completed')

        Label(self.master,
              text='Training dataset contains following features: ').grid(
                  row=self.row, column=1)
        self.row = self.row + 1
        Label(self.master, text=str(list(self.df_train.columns))).grid(
            row=self.row, column=1)
        self.row = self.row + 1
        self.master.update()

        self.logger.info('CTraining dataset contains the following features:')
        self.logger.info(str(list(self.df_train.columns)))

        if not self.from_scratch and len(ind) > 0:
            self.logger.warning(
                'Careful! Ratio of absence to presence data might be obscured')
            Label(self.master, text='Careful! Ratio might be obscured!').grid(
                row=self.row, column=1)
            self.row = self.row + 1
            self.master.update()

    def ensure_same_ratio(self):

        self.logger.info(
            'Process to ensure desired ratio\
                of absence to presence data started')

        len_pres = np.shape(self.df_train[self.df_train['label'] == 1])[0]
        len_abs = np.shape(self.df_train[self.df_train['label'] == 0])[0]

        if self.properties_train['num_nonls'] == self.length_before_cleaning:
            if len_abs > len_pres:
                self.logger.info('Number of absence locations is reduced')
                df_abs = self.df_train[self.df_train['label'] == 0]

                df_abs = df_abs.iloc[:len_pres]

                self.df_train = pd.concat(
                    [self.df_train[self.df_train['label'] == 1], df_abs],
                    axis=0)
            elif len_abs < len_pres:
                self.logger.error(
                    'Undefined error in the number\
                        of absence and presence data')

            Label(self.master,
                  text='Same ratio of absence to presence is ensured').grid(
                      row=self.row, column=1)
            self.logger.info(
                'Same ratio of absence and presence locations ensured')
            self.row = self.row+1
            self.master.update()
        else:
            if len_abs > self.properties_train['num_nonls']:
                df_abs = self.df_train[self.df_train['label'] == 0]

                df_abs = df_abs.iloc[:len_abs
                                     - (len_abs
                                        - self.properties_train['num_nonls'])]

                self.df_train = pd.concat(
                    [self.df_train[self.df_train['label'] == 1], df_abs],
                    axis=0)

                Label(self.master,
                      text='Number of absence locations is fixed').grid(
                          row=self.row, column=1)
                self.row = self.row + 1
                self.master.update()

                self.logger.info(
                    'Number of absence locations is set to the desired number')
                
    
                

    def cluster_landslide_locations(self, re=False, num=None):

        """
            Cluster the landslide locations.
            If clusters are too large this functions reclusters
            these clusters into smaller ones.

            Input:
                re: boolean, True if reclustering
                num: list, numbers of clusters to be reclustered

            Output:
                re_cluster_name: list, clusters for every entry
        """

        if re:
            self.logger.info('Landslide location reclustering started')
            re_cluster_name = []
            count, len_org = 0, len(self.bb)

            # Number of clusters to split too large cluster into
            num_cluster = 4
            for i in num:
                df_sub = self.df_train[self.df_train.cluster == i].copy()
                # Clustering
                kmeans = KMeans(init="random",
                                n_clusters=num_cluster,
                                n_init=10,
                                max_iter=300,
                                random_state=42)
                kmeans.fit(np.column_stack((list(df_sub['xcoord']),
                                            list(df_sub['ycoord']))))
                tmp = kmeans.labels_[:]

                # Rename clusters to fit into naming convention
                for c, j in enumerate(tmp):
                    if j == 0:
                        tmp[c] = i
                    else:
                        tmp[c] = len_org + count*(num_cluster-1) + (j-1)

                df_sub['cluster'] = tmp
                self.df_train = pd.concat(
                    [self.df_train[self.df_train.cluster != i], df_sub],
                    axis=0)
                # Store cluster numbers to be returned
                re_cluster_name.append(set(tmp))
                count = count + 1

            return re_cluster_name

        else:
            self.logger.info('Landslide location clustering started')
            # Clustering
            kmeans = KMeans(init="random",
                            n_clusters=self.num_clusters,
                            n_init=10,
                            max_iter=300,
                            random_state=42)
            kmeans.fit(np.column_stack((list(self.df_train['xcoord']),
                                        list(self.df_train['ycoord']))))
            self.df_train['cluster'] = kmeans.labels_[:]

        Label(self.master, text="Landslide locations were clustered").grid(
            row=self.row, column=1)
        self.row = self.row + 1
        self.master.update()

        self.logger.info('Landslide locations clustered')

    def determine_bb_for_clustering(self, re=False, re_num=None):

        """
            Determine bounding box for the individual clusters.

            Input:
                re: boolean, True if reclustering
                num: list, numbers of clusters to be reclustered
        """

        if self.cluster:
            # Initial clustering
            self.logger.info('Bounding box definition for clusters started')
            if not re:
                self.bb = []
                if self.from_scratch:
                    to_enter = range(self.num_clusters)
                else:
                    # When adding a feature
                    if 'cluster' in self.df_train:
                        to_enter = list(set(self.df_train.cluster.to_list()))
                    else:
                        to_enter = range(self.num_clusters)
                for num in to_enter:
                    df_tmp = self.df_train[
                        self.df_train['cluster'] == num].copy()
                    df_tmp = df_tmp.reset_index(drop=True)
                    max_x = df_tmp['xcoord'].loc[df_tmp['xcoord'].idxmax()]
                    min_y = df_tmp['ycoord'].loc[df_tmp['ycoord'].idxmin()]
                    max_y = df_tmp['ycoord'].loc[df_tmp['ycoord'].idxmax()]
                    min_x = df_tmp['xcoord'].loc[df_tmp['xcoord'].idxmin()]

                    self.bb.append([max_y, min_y, min_x, max_x])
            # Reclustering
            else:
                bb_new = [0 for i in range(
                    len(set(self.df_train.cluster.to_list())))]
                bb_new[:len(self.bb)] = self.bb

                for num in re_num:

                    df_tmp = self.df_train[
                        self.df_train['cluster'] == num].copy()
                    df_tmp = df_tmp.reset_index(drop=True)
                    max_x = df_tmp['xcoord'].loc[df_tmp['xcoord'].idxmax()]
                    min_y = df_tmp['ycoord'].loc[df_tmp['ycoord'].idxmin()]
                    max_y = df_tmp['ycoord'].loc[df_tmp['ycoord'].idxmax()]
                    min_x = df_tmp['xcoord'].loc[df_tmp['xcoord'].idxmin()]

                    # Make sure that the order is
                    # preserved to match bounding boxes properly
                    if num >= self.num_clusters:
                        bb_new[num] = [max_y, min_y, min_x, max_x]
                    else:
                        bb_new[num] = [max_y, min_y, min_x, max_x]

                self.bb = bb_new.copy()

        else:

            self.logger.info('Bounding box definition started')

            max_x = self.df_train['xcoord'].loc[
                self.df_train['xcoord'].idxmax()]
            min_y = self.df_train['ycoord'].loc[
                self.df_train['ycoord'].idxmin()]
            max_y = self.df_train['ycoord'].loc[
                self.df_train['ycoord'].idxmax()]
            min_x = self.df_train['xcoord'].loc[
                self.df_train['xcoord'].idxmin()]

            self.bb = [max_y, min_y, min_x, max_x]
            print(self.bb)

        Label(self.master, text="Bounding boxes were determined").grid(
            row=self.row, column=1)
        self.row = self.row + 1
        self.master.update()

        self.logger.info('Bounding boxes defined')

    def determine_if_reclustering(self):

        """
            Determine if the extent of one or several
            clusters are too large for local interpolation

            Output:
                num_bb: list, names of clusters that need reclustering
        """

        self.reclustering = False
        num_bb = []

        self.logger.info('Determine if reclustering is necessary')

        # Check extend of individual clusters
        for count, bb in enumerate(self.bb):
            point1_x = LatLon(Latitude(bb[0]), Longitude(bb[2]))
            point2_x = LatLon(Latitude(bb[0]), Longitude(bb[3]))
            distance_x = point1_x.distance(point2_x)*1000

            point1_y = LatLon(Latitude(bb[0]), Longitude(bb[2]))
            point2_y = LatLon(Latitude(bb[-1]), Longitude(bb[2]))
            distance_y = point1_y.distance(point2_y)*1000

            num_px_x = int(np.round(
                (distance_x/self.properties_settings['resolution'])))
            num_px_y = int(np.round(
                (distance_y/self.properties_settings['resolution'])))

            if num_px_x or num_px_y > 10000:
                num_bb.append(count)
                self.reclustering = True
                self.logger.info('Reclustering necessary')
            else:
                self.logger.info('Reclustering not necessary')

        return num_bb

    def open_secondary_window(self, kind, maximum):

        # Create secondary (or popup) window.
        self.secondary_window = tk.Toplevel()
        self.secondary_window.title("Progress")
        self.secondary_window.config(width=300, height=200)
        if kind == 'string':
            self.progress_var = tk.StringVar()
        else:
            self.progress_var = tk.IntVar()
        self.label = Label(self.secondary_window, text="")
        self.label.grid(row=1, column=0)
        progressbar = ttk.Progressbar(self.secondary_window,
                                      variable=self.progress_var,
                                      maximum=maximum)
        progressbar.grid(row=0, column=0)

        # Create a button to close (destroy) this window.
        button_close = ttk.Button(
            self.secondary_window,
            text="Close window",
            command=self.secondary_window.destroy
        )
        button_close.grid(row=2, column=0)

        self.secondary_window.focus()
        self.secondary_window.grab_set()

    def main_cluster(self):

        """
            Main function to generate training
            dataset if training locations shall be clustered.
        """

        def extract_data_from_dataset_subsets(num):

            """
                Extract information from the interpolated geospatial dataset.

                Input:
                    num: int, number of cluster

                Output:
                    df_clust: dataframe, subset of the training dataset
                    supplemented with the information from the training dataset
            """

            df_clust = self.df_train[self.df_train.cluster == num].copy()

            # Access interpolated subset of the dataset
            arr = self.ds['Result' + str(num)][:, :].data

            # Access interpolation x and y vector
            xvector = self.ds['Longitude' + str(num)][:].data
            yvector = self.ds['Latitude' + str(num)][:].data

            # Extract information at training location
            feature = []
            for index, row in df_clust.iterrows():

                x_indx = int((np.abs(np.array(xvector)
                                     - row['xcoord'])).argmin())
                y_indx = int((np.abs(np.array(yvector)
                                     - row['ycoord'])).argmin())

                feature.append(arr[y_indx, x_indx])

            df_clust[dataset] = feature

            return df_clust

        self.logger.info('Training dataset is generated\
                         using clustering and interpolation')

        Label(self.master,
              text='Training dataset is generated using\
                  clustering and interpolation').grid(row=self.row, column=1)
        self.row = self.row + 1
        self.master.update()

        # If previous run failed dataset might already exist causing
        # current run to crash again therefore it is deleted if it exists
        if os.path.isfile('tmp.nc'):
            os.remove('tmp.nc')

        if self.from_scratch:
            self.logger.info('Training dataset is generated from scratch')

            if not isinstance(self.df_train, pd.DataFrame):
                self.import_landslide_database()  # Import landslide database
            self.cluster_landslide_locations()

        if not self.from_scratch and 'cluster' not in self.df_train:
            # Cluster the landslide locations
            self.cluster_landslide_locations()
        # Determine the bounding boxes for the individual clusters
        self.determine_bb_for_clustering()

        if self.from_scratch:
            # Determine if bounding boxes are too large for local interpolation
            num_bb = self.determine_if_reclustering()

            if self.reclustering:
                Label(self.master, text="Reclustering necessary").grid(
                    row=self.row, column=1)
                self.row = self.row + 1
                self.master.update()
                re_cluster_name = self.cluster_landslide_locations(True,
                                                                   num_bb)
                re_cluster_name = [item for sublist in re_cluster_name
                                   for item in sublist]
                # Determine bounding boxes for new clusters
                self.determine_bb_for_clustering(True, re_cluster_name)

        dataset_counter = 0
        # Iterate over the dataset to inclue in the training dataset
        k = 0
        self.open_secondary_window(kind='string',
                                   maximum=len(self.data_to_handle))
        self.logger.info('Iteration over datasets started')
        for dataset in tqdm(self.data_to_handle):

            self.logger.info('Dataset: ' + dataset)

            self.progress_var.set(k)
            self.label.config(text=dataset)
            self.master.update()
            k += 1

            index = self.datasets_summary['keys'].tolist().index(dataset)
            # Call cut_and_interpolate class to cut and
            # interpolate the current dataset
            self.logger.info('Cutting and interpolation started')
            s = cut_and_interpolate(key=dataset,
                                    path=self.datasets_summary[
                                        'path'].tolist()[index],
                                    no_data_value=self.datasets_summary[
                                        'no_value'].tolist()[index],
                                    categorical=self.datasets_summary[
                                        'categorical'].tolist()[index],
                                    several=False,
                                    several_same=False,
                                    first=False,
                                    bb=self.bb,
                                    cluster=self.cluster,
                                    prop_settings=self.properties_settings,
                                    prop=self.properties_train,
                                    path_properties='properties.pkl')
            self.logger.info('Cutting and interpolation completed')
            # Open the netcdf file which contains the interpolated subsets of
            # the dataset with the extent of the bounding boxes of the clusters
            self.ds = nc.Dataset('tmp.nc')

            # Determine if one or more bounding boxes is
            # outside of the extend of the dataset
            if not s.cuttable:
                self.logger.error(
                    'Bounding box larger than dataset!\
                        Please adapt bounding box!')
                break

            df = []
            # Iterate over the clusters extract information from the dataset
            self.logger.info('Extraction of gridded information started')
            for num in range(len(self.bb)):
                df.append(extract_data_from_dataset_subsets(num))
            self.logger.info('Extraction of gridded information completed')

            # Concatenate all subsets of the training dataset
            df = np.concatenate(df, axis=0)

            # For first dataset set up final training dataset,
            # for later datasets append information
            if dataset_counter == 0:
                df_ges = pd.DataFrame(df, columns=list(
                    self.df_train.columns)+[dataset])
            else:
                df_ges[dataset] = df[:, -1]

            dataset_counter = dataset_counter + 1
            self.ds.close()

        self.progress_var.set(k)
        self.master.update()

        self.label.config(text='All features have been successfully created')
        Label(self.master, text='All datasets have been handled').grid(
            row=self.row, column=1)
        self.row = self.row + 1
        self.master.update()

        self.logger.info('All features have been successfully created')

        # Remove temporary netcdf file that
        # contains interpolated subsets of the dataset
        self.df_train = df_ges.copy()
        self.df_train.reset_index(inplace=True, drop=True)

        self.clean_df()  # Clean the DataFrame from no value rows
        if not self.from_scratch and not self.delete:
            self.handle_categorical_features(var=self.data_to_handle)
        else:
            self.handle_categorical_features()
            
        if self.from_scratch:
            # Ensure that the 1:1 ratio of the training dataset is perserved
            self.ensure_same_ratio()

        self.save_training_data()   # Save the training dataset as csv file

    def main(self):

        self.logger.info('Training dataset is generated using interpolation')

        Label(self.master,
              text="Training dataset is generated using interpolation").grid(
                  row=self.row, column=1)
        self.row = self.row + 1
        self.master.update()

        self.import_landslide_database()
        self.determine_bb_for_clustering()

        self.logger.warning('Warning! Depending on the size of the area of\
                            interest and the set resolution this might be\
                            computationally expensive. Consider clustering.')

        Label(self.master,
              text='Warning! Depending on the size of the area of\
                  interest and the set resolution this might be\
                      computationally expensive. Consider clustering.').grid(
                      row=self.row, column=1)
        self.row = self.row + 1
        self.master.update()

        self.logger.info('Generation if data cube started')
        self.s = generate_data_matrix(from_scratch=True,
                                      delete=False,
                                      dataset='training',
                                      bb=self.bb,
                                      data_to_handle=self.data_to_handle,
                                      geo_overview=self.overview,
                                      settings_train_pred=self.properties_train,
                                      settings=self.properties_settings)
        Label(self.master, text="Data matrix was generated").grid(
            row=self.row, column=1)
        self.row = self.row + 1
        self.master.update()
        self.logger.info('Data cube generated')

        self.import_cube()
        # Extract geospatial information at ls and non-ls locations
        self.extract_gridded_info()

        # Concat final training dataset
        self.df_train = pd.concat([self.df_train, self.df_features], axis=1)

        self.clean_df()  # Clean the DataFrame from no value rows
        self.handle_categorical_features()
        self.ensure_same_ratio()
        self.save_training_data()   # Save the training dataset as csv file

    def import_raw_dataset(self, path, no_data):

        self.logger.info('Importing raw dataset')
        if path.split('.')[-1] == 'tif':
            data, x, y, _ = import_tif(path)
        elif path.split('.')[-1] == 'nc':
            data, x, y, _ = import_nc(path)
        else:
            self.logger.warning(
                'Not the right data format! Please provide tif or nc file')

        if y[0] < y[-1]:
            y = np.flip(y)
            
        if no_data != 'None':
            no_data = no_data.split(',')
            no_data = [float(val) for val in no_data]
        else:
            no_data = no_data
        if no_data != 'None':
            for val in no_data:
                data[data == val] = self.properties_settings['no_value']

        self.logger.info('Raw dataset imported')

        return data, x, y

    def main_no_interpolation(self):

        Label(self.master,
              text="Training dataset generation\
                  is done without interpolation").grid(row=self.row, column=1)
        self.row = self.row + 1
        self.master.update()

        self.logger.info('Training dataset generated without interpolation')

        def extract_gridded_info(row):

            x_indx = int((np.abs(np.array(x) - row['xcoord'])).argmin())
            y_indx = int((np.abs(np.array(y) - row['ycoord'])).argmin())

            return data[y_indx, x_indx]

        if self.from_scratch:
            self.logger.info('Training dataset is generated from scratch')
            self.import_landslide_database()

        k = 0
        self.open_secondary_window(kind='string',
                                   maximum=len(self.data_to_handle))
        self.logger.info('Iterate over datasets:')
        for dataset in tqdm(self.data_to_handle):
            self.logger.info('Dataset: ' + dataset)
            self.progress_var.set(k)
            self.label.config(text=dataset)
            k += 1

            index = self.datasets_summary['keys'].tolist().index(dataset)
            data, x, y = self.import_raw_dataset(
                self.datasets_summary['path'].tolist()[index],
                self.datasets_summary['no_value'].tolist()[index])
            feat = []
            self.logger.info('Extraction of gridded information started')
            for index, row in self.df_train.iterrows():
                feat.append(extract_gridded_info(row))
            self.logger.info('Extraction of gridded information completed')

            self.df_train[dataset] = feat
            self.master.update()
        self.progress_var.set(k)
        self.master.update()

        self.label.config(text='All features have been successfully created')
        Label(self.master, text='All datasets have been handled').grid(
            row=self.row, column=1)
        self.row = self.row + 1
        self.master.update()


        self.clean_df()  # Clean the DataFrame from no value rows
        if not self.from_scratch and not self.delete:
            self.handle_categorical_features(var=self.data_to_handle)
        else:
            self.handle_categorical_features()
            self.ensure_same_ratio()
        self.save_training_data()
            
    def handle_categorical_features(self, var=None):
        
        """
            Function is called which performs one-hot or ordinal encoding
        """
        
        basic = ['ID', 'xcoord', 'ycoord', 'label']
        self.df_train = handle_categorical_values(self.df_train,
                                                  self.datasets_summary,
                                                  self.properties_train['ohe'],
                                                  basic,
                                                  var)

    def save_training_data(self):

        """
            Save dataframe as csv. If necessary folder is created.
        """

        self.logger.info('Saving of training data in progress')

        outfile = self.properties_train['train_path']

        isExist = os.path.dirname(self.properties_train['train_path'])
        if not isExist:
            os.path.dirname(self.properties_train['train_path'])

        # If outfile exists already, delete
        if os.path.exists(outfile):
            os.remove(outfile)

        # Save dataframe as csv
        self.df_train.to_csv(outfile, sep=',', index=False)

        self.logger.info('Training dataset saved')

        Label(self.master, text='Training dataset has been saved').grid(
            row=self.row, column=1)
        self.row = self.row + 1
        self.master.update()

        button_close = ttk.Button(
            self.master,
            text="Close window",
            command=self.master.destroy
        )
        button_close.grid(row=self.row, column=0)
        self.master.update()
