#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy as np
import pandas as pd
import netCDF4 as nc
import pickle as pkl
import os
import pickle

from sklearn.model_selection import train_test_split
from sklearn.ensemble import RandomForestClassifier
from sklearn.metrics import mean_squared_error, f1_score, roc_curve, auc, fbeta_score
from joblib import Parallel, delayed

import settings
from utilities.ncfile_generation import generate_ncfile
from utilities.strings_for_ncfile import char_to_string


class prepare_data:

    """
        This class prepares the data to be
        used in the Random Forest classifier.
    """

    def __init__(self, aim, logger, retrain):

        invalid = False
        self.aim = aim
        self.logger = logger
        self.retrain = retrain
        if aim == 'train_test':
            print('Train the model')
            invalid = False
        elif aim == 'prediction':
            print('Prepare the hazard map')
            invalid = False
        else:
            print('Not a valid command. Enter train_test or prediction.')
            invalid = True

        if not invalid:
            self.test_size = settings.size  # Size of validation dataset
            # Column name of label in training dataset
            self.label_name = 'label'
            self.xy = pd.DataFrame()  # Array to save coordinates for reshaping

            if aim == 'train_test':
                self.import_features_labels()  # Prepare the training
                # Generate train/validation dataset
                self.split_training_testing()
            elif aim == 'prediction':
                self.import_features()  # Import prediction dataset
        
    def import_features(self):

        """
            Imports the features for prediction.
        """

        # Import prediction dataset either as csv file or nc file
        
        if os.path.isdir(settings.path_pred):
            path_pred = settings.path_pred + 'prediction.nc'
        else:
            path_pred = settings.path_pred
        
        if path_pred.split('.')[-1] == 'csv':
            self.features = pd.read_csv(path_pred)

        elif path_pred.split('.')[-1] == 'nc':
            ds = nc.Dataset(path_pred)
            pred = ds['Result'][:, :].data
            pred_features = ds['features'][:].data
            self.feature_list = char_to_string(pred_features)

            if 'xcoord' in self.feature_list and 'ycoord' in self.feature_list:
                self.features = pd.DataFrame(pred, columns=self.feature_list)
            else:
                self.features = pd.DataFrame(pred, columns=['xcoord', 'ycoord']+self.feature_list)
                
            self.dropped = ds['Dropped'][:].data
            self.dropped = [int(x) for x in self.dropped]
            
        # Save the prediction coordinates in the prediction dataset
        self.xy['ycoord'] = self.features['ycoord']
        self.xy['xcoord'] = self.features['xcoord']

        # Remove all features that shall not be included in
        # prediction from DataFrame (see settings!)
        if len(settings.not_included_pred_data) > 0:
            for dataset in settings.not_included_pred_data:
                self.features = self.features.drop(dataset, axis=1)
                
        # Determine which classes are contained in the categorical features
        # It is distinguished between one-hot and ordinal encoded features
        self.categorical_classes = {}
        cat_subset = [feat for feat in self.features.columns.tolist() if '_encoded' in feat]
        df_sub = self.features[cat_subset]
        cat_feat = ['_'.join(col.split('_')[:len(col.split('_'))-1]) for col in df_sub.columns.tolist()]
        self.distibuish_encoding = {}
        for feat in list(set(cat_feat)):
            classes = []
            if cat_feat.count(feat)>1:
                classes.append([f.split('_')[-2] for f in df_sub.columns.tolist() if feat in f])
                self.distibuish_encoding[feat] = 'ohe'
            else:
                classes.append(list(set(df_sub[feat + '_encoded'].tolist())))
                self.distibuish_encoding[feat] = 'ordinal'
            self.categorical_classes[feat] = {}
            self.categorical_classes[feat]['classes'] = [item for sublist in classes for item in sublist]
            self.categorical_classes[feat]['num_cols'] = cat_feat.count(feat)
            
        self.feature_list = list(self.features.columns)
        self.features_org = self.features.copy()

        self.logger.info('Features for prediction were imported')
        self.logger.info('The following '
              + str(len(self.feature_list))
              + ' features are included in the prediction dataset: '
              + str(self.feature_list))

    def import_features_labels(self):

        """
            Imports the features for training.
        """

        # Import training dataset as csv file
        if os.path.isdir(settings.path_train):
            self.features = pd.read_csv(settings.path_train + 'training.csv')
        else:
            self.features = pd.read_csv(settings.path_train)
        # Extract and remove labels from training dataset
        self.labels = np.array(self.features[self.label_name]).reshape(
            [np.shape(self.features[self.label_name])[0], 1])
        self.features = self.features.drop(self.label_name, axis=1)

        # Store coordinates from training data
        self.xy['ycoord'] = self.features['ycoord']
        self.xy['xcoord'] = self.features['xcoord']

        # Drop ID from training data
        self.features = self.features.drop('ID', axis=1)
        self.features = self.features.drop(['xcoord', 'ycoord'], axis=1)

        # Remove all features that shall not be included in
        # training from DataFrame (see settings!)
        
        if self.retrain:
            features_to_remove = pd.read_csv(settings.path_ml + settings.model_to_save + '/feature_mismatch_training.csv')['to_drop'].to_list()
            not_included_train_data = settings.not_included_train_data + features_to_remove
        else:
            not_included_train_data = settings.not_included_train_data
            
        
        if len(not_included_train_data) > 0:
            for dataset in not_included_train_data:
                self.features = self.features.drop(dataset, axis=1)
                
        # Determine which classes are contained in the categorical features
        # It is distinguished between one-hot and ordinal encoded features
        self.categorical_classes = {}
        cat_subset = [feat for feat in self.features.columns.tolist() if '_encoded' in feat]
        df_sub = self.features[cat_subset]
        cat_feat = ['_'.join(col.split('_')[:-2]) for col in df_sub.columns.tolist()]
        for feat in list(set(cat_feat)):
            classes = []
            if cat_feat.count(feat)>1:
                classes.append([f.split('_')[-2] for f in df_sub.columns.tolist() if feat in f])
            else:
                classes.append(list(set(df_sub[feat + '_encoded'].tolist())))
            self.categorical_classes[feat] = {}
            self.categorical_classes[feat]['classes'] = [item for sublist in classes for item in sublist]
            self.categorical_classes[feat]['num_cols'] = cat_feat.count(feat)
        

        self.feature_list = list(self.features.columns)
        self.logger.info('Features for training were imported')
        self.logger.info('The following ' + str(len(self.feature_list))
              + ' features are included in the training dataset: '
              + str(self.feature_list))
        self.features = np.array(self.features)

    def split_training_testing(self):

        """
            Splits the training data into training and validation data.
        """

        self.train_features, self.test_features, self.train_labels, self.test_labels = \
            train_test_split(self.features,
                             self.labels,
                             test_size=self.test_size,
                             random_state=settings.random_seed,
                             stratify=self.labels)
        print('Data split')
        self.logger.info('Training data split in training and test dataset')


class RandomForest(prepare_data):

    def __init__(self, aim, parallel=False, log=None, retrain=None):

        super().__init__(aim, log, retrain)

        self.aim = aim
        self.parallel = parallel
        self.retrain = retrain
        self.logger = log
        self.num_chunks = 10
        # Random Forest settings
        self.criterion = settings.criterion
        self.n_estimators = settings.num_trees
        self.max_depth = settings.depth

        self.model_dir = settings.model_database_dir
        if self.retrain:
            self.model_to_load = settings.model_to_load + '_retrain'
            self.model_to_save = settings.model_to_save + '_retrain'
        else:
            self.model_to_load = settings.model_to_load
            self.model_to_save = settings.model_to_save
        self.output_dir = None

        if aim == 'train_test':
            print('Model is trained')
            self.define()
            self.train()
            self.predict()
            self.evaluate()
            self.create_output_dir()
            self.save_model()
            self.save_parameters()
            self.feature_importance()

        elif aim == 'prediction':
            print('Prediction is performed')
            self.create_output_dir()
            self.load_model()
            if not self.error:
                self.predict()
                self.extract_pos_neg_predictions()
                self.reshape_prediction()
                self.save_prediction()

    def define(self):

        """
            Define the Random Forest Classifier model.
        """

        self.model = RandomForestClassifier(n_estimators=self.n_estimators,
                                            max_depth=self.max_depth,
                                            random_state=settings.random_seed)
        self.logger.info('Model is defined')

    def train(self):

        """
            Train the Random Forest Classifier model.
        """

        self.model.fit(self.train_features, np.ravel(self.train_labels))
        self.logger.info('Model is trained')

    def predict(self):

        """
            Make the prediction.
        """

        print('Predicting...')
        self.logger.info('Predicting...')
        if self.aim == 'prediction':
            pred = self.features
        elif self.aim == 'train_test':
            pred = self.test_features

        if self.parallel:
            self.split_array_into_chunks(pred)
            prediction = Parallel(n_jobs=10)(
                delayed(self.model.predict)(chunk) for chunk in self.chunks)
            self.prediction = np.concatenate(prediction, axis=0)
        else:
            self.prediction = self.model.predict(pred)

    def split_array_into_chunks(self, pred):

        """
        Split a NumPy array into chunks without changing the number of columns.

        """
        self.logger.info('Prediction dataset is split in chunks')
        # Calculate the number of rows in each chunk
        rows_per_chunk = pred.shape[0] // self.num_chunks
        remaining_rows = pred.shape[0] % self.num_chunks

        # Create chunks
        self.chunks = []
        start = 0
        length = 0
        for i in range(self.num_chunks):
            end = start + rows_per_chunk + (1 if i < remaining_rows else 0)
            chunk = pred[start:end, :]
            self.chunks.append(chunk)
            start = end
            length = length + len(chunk)

    def evaluate(self):

        """
            Evaluate the validation dataset.
        """
        
        self.logger.info('Model is evaluated')
        y_pred_prob = self.model.predict_proba(self.test_features)[:, 1]
        self.fpr, self.tpr, self.thresholds = roc_curve(self.test_labels, y_pred_prob)

        # Calculate AUC (Area Under the Curve)
        self.roc_auc = auc(self.fpr, self.tpr)

        diff = [abs(pred-test_labels) for pred, test_labels
                in zip(list(self.prediction), list(self.test_labels))]
        self.acc = str(diff.count(1)) + '/' +  str(len(diff))
        self.mae = round(np.mean(diff), 2)
        print('Mean absolute error: ' + str(self.mae))
        print('Wrongly predicted: '
              + str(np.count_nonzero(diff))
              + '/' + str(len(diff)))
        self.mse = mean_squared_error(self.test_labels, self.prediction)
        self.f1 = f1_score(self.test_labels, self.prediction)
        self.fbeta = fbeta_score(self.test_labels, self.prediction, beta=2)
        print('Mean squared error: ' + str(self.mse))

    def create_output_dir(self):

        """
            Define and create the output directory.
        """

        self.output_dir = self.model_dir + self.model_to_save

        if not os.path.isdir(self.output_dir):
            os.mkdir(self.output_dir)

    def save_model(self):

        """
            Save the Random Forest Classifier model.
        """

        with open(self.output_dir + '/saved_model.pkl', 'wb') as file:
            pkl.dump(self.model, file)
        self.logger.info('Model is saved')

    def save_parameters(self):

        """
            Save the metadata associated with the prediction.
        """

        params = {'Area': settings.bounding_box,
                  'criterion': [self.criterion],
                  'n_estimators': [self.n_estimators],
                  'max_depth': [self.max_depth],
                  'features': self.feature_list,
                  'mse': self.mse,
                  'mae': self.mae,
                  'f1': self.f1,
                  'roc_threshold': self.thresholds,
                  'roc_fpr': self.fpr, 
                  'roc_tpr': self.tpr,
                  'roc_auc': self.roc_auc,
                  'accuracy': self.acc,
                  'fbeta': self.fbeta,
                  'categories': self.categorical_classes
                  }

        with open(settings.model_database_dir
                  + self.model_to_save
                  + '/model_params.pkl', 'wb') as file:
            pkl.dump(params, file)
            
        self.logger.info('Parameters are saved')
        
    def adapt_categorical_features(self, train_classes, training_features):
        
        """
            The encoded features in the training and prediction dataset are
            compared regarding the contained classes. Depending on the user
            input, instances in the prediction dataset with classes that are
            not included in the training dataset are either set to no_value or
            nevertheless considered in the prediction. The surplus additional
            features are removed either way to achieve the same set of features
            as in the training dataset
        """
        
        self.instances_to_drop = []
        self.features_not_in_training = []
        
        for feat in [val for val in training_features if '_encode' in val]:
            if feat not in self.feature_list:
                print('Error: cannot proceed with mapping')
                print('Error: Categorical feature ' + feat + ' not in prediction dataset')
                self.logger.error('Error: Categorical feature ' + feat + ' not in prediction dataset')
                
                self.error = True
                self.retrain = True
                self.features_not_in_training.append(feat)
                
        if len(self.features_not_in_training) > 0:
                pd.DataFrame(self.features_not_in_training, columns=['to_drop']).to_csv(self.model_dir + self.model_to_load + 'feature_mismatch_training.csv', index=False)

        if not self.retrain:
            if list(set([val for val in training_features if '_encode' in val])) != list(set(self.feature_list)):    
                for feat in list(set(['_'.join(val.split('_')[:-2]) for val in self.feature_list if '_encode' in val])):
                    if feat in list(self.distibuish_encoding.keys()):
                        if self.distibuish_encoding[feat] == 'ohe':
                            if (train_classes[feat]['num_cols'] < self.categorical_classes[feat]['num_cols']) or (set(train_classes[feat]['classes']) != set(self.categorical_classes[feat]['classes'])):
                                print(feat + ': Prediction dataset contains more or other classes than training dataset')
                                
                                self.logger.warning(feat + ': Prediction dataset contains more classes than training dataset')
                                self.logger.info('Apply user defined handling approach')
                                
                                common_elements = set(train_classes[feat]['classes']).intersection(set(self.categorical_classes[feat]['classes']))
                            
                                if self.properties_map['keep']:
                                    if len(common_elements) == 0:
                                        print('Error: no common classes for ' + feat + ' in training and prediction dataset')
                                        self.logger.error('Error: no common classes for ' + feat + ' in training and prediction dataset')
                                        self.error = True
                                    else:
                                        to_drop = [feat + '_' + str(f) + '_encode' for f in self.categorical_classes[feat]['classes'] if f not in common_elements]
                                        self.features = self.features.drop(to_drop, axis=1)
                                        self.feature_list = self.features.columns.tolist()
                                elif self.properties_map['remove_instances']:
                                    to_drop_col = [feat + '_' + str(f) + '_encode' for f in self.categorical_classes[feat]['classes'] if f not in common_elements]
                                    to_drop_row = []
                                    for col in to_drop_col:
                                        to_drop_row = to_drop_row + self.features.index[self.features[col] == 1].tolist()
                                    self.features = self.features.drop(to_drop_col, axis=1)
                                    
                                    print('Not matching features have been removed')
                                    self.logger.info('Not matching features have been removed')
                                    
                                    self.feature_list = self.features.columns.tolist()
                                    self.instances_to_drop = self.instances_to_drop + to_drop_row
                                    
                                    print('Instances to consider during mapping have been adapted')
                                    self.logger.info('Instances to consider during mapping have been adapted')

            print('Categorical features have been handled and hamonised')
            self.logger.info('Categorical features have been handled and hamonised')
            self.logger.info('Remaining features: ' + str(self.feature_list))
                        

    def load_model(self):

        """
            Load the Random Forest Classifier model and the metadata.
            Make sure to compare features of training and prediction dataset
            as well as their order.
        """                    
                
        print('Loading model ' + self.model_dir
              + self.model_to_load + '/saved_model.pkl')
        self.logger.info('Loading model ' + self.model_dir
              + self.model_to_load + '/saved_model.pkl')
        with open(self.model_dir
                  + self.model_to_load + '/saved_model.pkl', 'rb') as file:
            self.model = pkl.load(file)

        with open(settings.model_database_dir
                  + self.model_to_save
                  + '/model_params.pkl', 'rb') as f:
            params = pkl.load(f)
        features = params['features']
        self.error = False
        self.adapt_categorical_features(params['categories'], features)

        if not self.error:
            if len(self.feature_list) == len(features):
                if set(self.feature_list) != set(features):
                    
                    print('Error: Not all features of the model are contained in the prediction dataset')
                    self.logger.error('Error: Not all features of the model are contained in the prediction dataset')
                    
                    self.error = True
                elif self.feature_list != features:
                    
                    print('The order or features differs. Prediction features are reordered')
                    self.logger.info('The order or features differs. Prediction features are reordered')
                    
                    self.features = self.features[features]
                    if self.features.columns.tolist() != features:
                        print('There is still something wrong with the order of the features!')
                       
                elif self.feature_list == features:
                    
                    print('Prediction and training dataset have the same order')
                    self.logger.info('Prediction and training dataset have the same order')
            elif len(self.feature_list) < len(features):
                
                print('Error: Not all features of the model are contained in the prediction dataset')
                self.logger.error('Error: Not all features of the model are contained in the prediction dataset')
                
                self.error = True
            elif len(self.feature_list) > len(features):
                if set(features).issubset(self.feature_list):
                    to_drop = list(set(self.feature_list)-set(features))
                    self.features = self.features.drop(to_drop, axis=1)
                    self.feature = self.features[features]
                    if self.features.columns.tolist() != features:
                        print('There is still something wrong with the order of the features!')
                        self.error = True
                    else:                      
                        print('Features in the prediction dataset which were not used for training were removed')
                        print('Features in the prediction dataset were sorted to match the training features')
                        
                        self.logger.warning('Features in the prediction dataset which were not used for training were removed')
                        self.logger.info('Features left: ' + str(self.feature_list))
                        self.logger.info('Features in the prediction dataset were sorted to match the training features')
                else:
                    Label(self.master, text='Error: Not all features of the model are contained in the prediction dataset').grid(
                                row=self.row, column=1)
                    self.row = self.row + 1
                    self.master.update()
                    
                    self.logger.error('Error: Not all features of the model are contained in the prediction dataset')
                    
                    self.error = True
            if not self.error:
                self.feature_list = self.features.columns.tolist()
                self.features = self.features.to_numpy()  
        
        self.logger.info('Model loaded from '
                         + self.model_dir
                         + self.model_to_load)
        
        print('Model loaded from '
                         + self.model_dir 
                         + self.model_to_load)
        print("Model successfully loaded")
        
    def save_prediction(self):

        """
            Save the prediction.
        """

        if self.aim == 'prediction':
            output_dir = self.model_dir + self.model_to_load

        self.xy.to_csv(output_dir + '/prediction_results.csv',
                       columns=['xcoord', 'ycoord', 'pred'],
                       index=True)
        self.df_pos.to_csv(output_dir + '/pos_prediction_results.csv',
                           columns=['xcoord', 'ycoord', 'pred'],
                           index=True)
        self.df_neg.to_csv(output_dir + '/neg_prediction_results.csv',
                           columns=['xcoord', 'ycoord', 'pred'],
                           index=True)
        print('Predictions saved in ' + output_dir)
        self.logger.info('Prediction saved in ' + output_dir)

    def reshape_prediction(self):

        """
            Reshape the individual predictions into a map.
        """

        dropped = list(set(self.dropped + self.instances_to_drop))
        arr_xy = np.array(self.xy)
        arr_xy[dropped, :] = settings.no_value#*np.shape(arr_xy)[1]

        result = np.reshape(list(arr_xy[:, 2]),
                            (len(list(set(self.xy['ycoord']))),
                             len(list(set(self.xy['xcoord'])))))
        self.logger.info('Prediction is reshaped into the final map')

        self.save_prediction_as_nc(result)

    def extract_pos_neg_predictions(self):

        """
            Distinguish between the classes of the Classifier.
        """

        print('Extract pos and neg predictions...')
        self.logger.info('Extract positive and negative predictions...')
        self.xy['pred'] = self.prediction
        self.df_pos = self.xy[self.xy.pred == 1]
        self.df_neg = self.xy[self.xy.pred == 0]

    def save_prediction_as_nc(self, prediction):

        """
            Save the hazard map to a netCDF4 file.
        """

        print('Saving as nc-File')
        outfile_name = self.model_dir + self.model_to_load + '/prediction.nc'

        if os.path.exists(outfile_name):
            os.remove(outfile_name)

        generate_ncfile(outfile_name,
                        np.array(list(set(self.xy['xcoord']))),
                        np.array(list(set(self.xy['ycoord']))),
                        prediction,
                        crs=settings.crs,
                        missing_value=settings.no_value)
        self.logger.info('Map is saved as nc-file')

    def feature_importance(self):

        """
            Access feature importance information from the Random Forest.
        """

        feature_imp = pd.Series(self.model.feature_importances_,
                                index=self.feature_list).sort_values(
                                    ascending=False)

        feature_imp.to_csv(self.model_dir
                           + settings.model_to_load
                           + '/feature_importance.csv')
        self.logger.info('Feature importance is saved')
