#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy as np
import pandas as pd
import netCDF4 as nc
import os
import pickle

from tqdm import tqdm

import settings
from data_preprocessing import generate_data_matrix
from utilities.ncfile_generation import generate_basic_ncfile
from utilities.strings_for_ncfile import char_to_string, features_to_char
from utilities.handle_categorical_values import handle_categorical_values


class create_prediction_data:

    """
        This class creates the prediction data
        for the Random Forest classifier.

        Input:
            from_scratch: boolean, True if prediction dataset should
                          be generated from scratch, otherwise false
            delete: True if dataset/feature should be deleted
                    from prediction dataset
                    False if dataset should be added to existing
                    prediction dataset
                    (careful: from_scratch needs to be False!)

        Output:
            netCDF4 file
    """

    def __init__(self, from_scratch=True, delete=False, log=None):

        self.from_scratch = from_scratch
        self.delete = delete
        self.bb = settings.bounding_box
        self.logger = log

        if self.from_scratch:
            self.logger.info('Prediction dataset is generated from scratch')
            self.s = generate_data_matrix(from_scratch=True,
                                          delete=False,
                                          dataset='prediction',
                                          bb=self.bb)
            self.import_cube()  # Import data cube
            if not self.no_dataset_found:
                # Add coordinate information to prediction
                # dataset for map generation
                self.add_coordinates()
                # Flatten data cube for efficient information extraction
                self.flatten_cube()
                # Determine no value instances in DataFrame
                self.clean_df()
                self.import_data_summary()
                self.handle_categorical_features()
                # Save prediction dataset
                self.save_as_nc()

        elif not self.from_scratch and not self.delete:
            self.logger.info(
                'Feature(s) will be added to the prediction dataset')
            # Import information on the geospatial datasets
            self.import_data_summary()
            # Import existing prediction dataset
            self.import_prediction_dataset()
            # Import data cube that contains cut and
            # interpolate dataset to be added
            self.import_cube()

            # Check if datasets to be added are contained in the data cube
            not_included = False
            for feature in self.data_to_handle:
                if feature not in self.features:
                    print(str(feature)
                          + ' not included in cube, it has to be added first')
                    not_included = True

            if not_included:
                print('not included')
                self.s = generate_data_matrix(
                    from_scratch=False,
                    delete=False,
                    dataset='prediction',
                    bb=self.bb,
                    data_to_handle=self.data_to_handle,
                    keys_already_included=self.features)
                self.import_cube()

            self.add_feature()  # Add feature
            # Save prediction dataset
            self.clean_df()
            self.handle_categorical_features(var=self.data_to_handle)
            self.save_as_nc()

        elif not self.from_scratch and self.delete:
            
            self.logger.info(
                'Feature(s) will be removed from the prediction dataset')
            # Import existing prediction dataset
            self.import_data_summary()
            self.import_prediction_dataset()
            self.delete_features()  # Delete features from prediction dataset
            # Save prediction dataset
            self.save_as_nc()
            
    def import_data_summary(self):
        """
            Import the information on the geospatial datasets and the 
            keys to include in the prediction dataset
        """
        
        self.data_properties = pd.read_csv(settings.data_summary_path)
        self.keys_to_include = pd.read_csv(settings.key_to_include_path)['keys_to_include'].tolist()
        self.data_to_handle = self.keys_to_include
        self.logger.info('Data summary in keys to include have been imported')
        
    def handle_categorical_features(self, var=None):
        
        """
            Function is called which performs one-hot or ordinal encoding
        """

        basic = ['xcoord', 'ycoord']
        self.df_pred = handle_categorical_values(self.df_pred,
                                                 self.data_properties,
                                                 settings.ohe,
                                                 basic,
                                                 var)
        
        to_drop = []
        for col in self.df_pred.columns.tolist():
            if str(settings.no_value) in col:
                to_drop.append(col)

        self.df_pred = self.df_pred.drop(to_drop, axis=1)

    def add_feature(self):

        """
            Add feature to the prediction dataset
        """

        for count, key in enumerate(self.data_to_handle):
            # Delete feature if it already exists in training dataset
            if key in self.df_pred.columns:
                print(
                    'Feature already exists in prediction dataset.\
                        Existing feature is deleted')
                self.logger.info('Feature already exists in prediction dataset.\
                        Existing feature is deleted')
                
                self.df_pred = self.df_pred.drop(key, axis=1)

                print('Adding ' + key + '...')
                self.logger.info('Adding ' + key)

            # Create empty DataFrame
            if count == 0:
                self.df_features = pd.DataFrame(
                    index=range(len(self.df_pred)),
                    columns=self.data_to_handle)

            data_flat = self.cube[:, :, self.features.index(key)].flatten()
            self.df_features[key] = data_flat

            # Combine old training dataset with additional features
        self.df_pred = pd.concat([self.df_pred, self.df_features], axis=1)
        # Adapt column order
        #self.df_pred = self.df_pred[ref]

        print('Prediction dataset contains following features: '
              + str(list(self.df_pred.columns)))
        self.logger.info('Prediction dataset contains the following features: '
                         + str(list(self.df_pred.columns)))

    def import_cube(self):

        """
            Import cut and interpolated data cube
            which was created in pre-processing.py
        """
        self.logger.info('Import cube with interpolated datasets')

        # Path to the stored data cube (see data_preprocessing.py)
        folder = settings.path_pred.rsplit('/', 1)[0]
        path = folder + '/data_combined_prediction_' + str(settings.resolution) + '.nc'

        # Check if path exists and import the cube
        # as well as list of datasets it contains
        if not os.path.exists(path):
            print('Error: Dataset not found!')
            self.logger.error('Error! Cube not found!')
            self.no_dataset_found = True
        else:
            self.no_dataset_found = False
            ds = nc.Dataset(path)
            self.cube = ds['Result'][:, :, :].data
            self.x = ds['Longitude'][:].data
            self.y = ds['Latitude'][:].data
            self.pred_features = ds['features'][:].data
            self.features = char_to_string(self.pred_features)

            print('Features included in cube: ' + str(self.features))
            self.logger.info('Features included in cube ' + str(self.features))

    def flatten_cube(self):

        """
            Flatten the individual datasets of the data cube
        """

        print('Flatten cube...')
        self.logger.info('Flatten cube...')
        # Go through all datasets in the data cube
        for i in tqdm(range(np.shape(self.cube)[2])):
            data = self.cube[:, :, i]
            data_flat = data.flatten()  # Flatten the dataset
            # Save it to the DataFrame
            self.df_pred[self.features[i]] = data_flat

    def add_coordinates(self):

        """
            Add coordinate for which the model shall
            make an prediction to the DataFrame.
        """

        print('Add coordinates...')
        self.logger.info('Adding coordinates...')
        self.df_pred = pd.DataFrame(columns=['xcoord', 'ycoord']
                                    + self.features)
        self.X, self.Y = np.meshgrid(self.x, self.y)

        data_flat = self.X.flatten()
        self.df_pred['xcoord'] = data_flat
        data_flat = self.Y.flatten()
        self.df_pred['ycoord'] = data_flat

    def clean_df(self):

        """
            Clean the DataFrame from rows with no data values
        """

        self.logger.info('Check prediction dataset for no value rows')
        # Go through the DataFrame column by column and
        # check for indices of no value data
        ind = []
        for feature in tqdm(list(self.df_pred.columns)):
            tmp = [i for i, x in
                   enumerate(self.df_pred[feature].tolist())
                   if x == settings.no_value]
            ind.append(tmp)

        # Remove duplicates
        self.idx = []
        for i in range(len(ind)):
            self.idx = self.idx + ind[i]
        self.idx = list(set(self.idx))

        # Save information on invalid locations so that they
        # can masked out during hazard map generation
        print(str(len(list(set(self.idx)))) + ' rows will be saved to be\
              handled during map generation')
        self.logger.info(str(len(list(set(self.idx)))) + ' rows will be saved to be\
              handled during map generation')
        
    def delete_features(self):

        """
            Delete feature from prediction dataset
        """

        to_drop = []
        for feat in self.data_to_handle:
            for col in self.df_pred.columns.tolist():
                if feat in col:
                    to_drop.append(col)
            
        self.df_pred.drop(columns=to_drop, inplace=True)

        print('Features now included in prediction dataset: '
              + str(list(self.df_pred.columns)))
        self.logger.info('Features now included in prediction dataset: '
              + str(list(self.df_pred.columns)))

    def import_prediction_dataset(self):

        """
            Import existing prediction dataset
        """
        
        if 'prediction.nc' in settings.path_pred:
            path = settings.path_pred
        else:
            path = settings.path_pred + 'prediction.nc'
            
        
        if not os.path.exists(path):
            print('Error: existing prediction dataset could not be found')
        else:
            ds = nc.Dataset(path)
            pred = ds['Result'][:, :].data
            pred_features = ds['features'][:].data
            self.features = char_to_string(pred_features)
            self.idx = ds['Dropped'][:].data
            if 'xcoord' in self.features and 'ycoord' in self.features:
                cols = self.features
            else:
                cols = ['xcoord', 'ycoord'] + self.features
            self.df_pred = pd.DataFrame(pred, columns=cols)

        print('Features included in dataset: ' + str(self.features))

    def save_as_nc(self):

        """
            Save prediction dataset and information on dropped rows as nc-file
        """

        df_pred = self.df_pred.to_numpy()
        self.char_features = features_to_char(self.df_pred.columns.tolist())
        
        if 'prediction.nc' not in settings.path_pred:
            outfile = settings.path_pred + 'prediction.nc'
        else:
            outfile = settings.path_pred

        isExist = os.path.exists(os.path.dirname(outfile))
        if not isExist:
            os.makedirs(os.path.dirname(outfile))

        ds = generate_basic_ncfile(outfile, crs=None)
        ds.createDimension('lat', (np.shape(df_pred)[0]))
        ds.createDimension('lon', (np.shape(df_pred)[1]))
        ds.createDimension('ix', (len(self.idx)))
        ds.createDimension('feat', len(self.char_features))
        result = ds.createVariable('Result', 'f4', ('lat', 'lon'))
        dropped = ds.createVariable('Dropped', 'f4', 'ix')
        Features = ds.createVariable('features', 'S1', 'feat')
        result[:, :] = df_pred
        dropped[:] = self.idx
        Features[:] = self.char_features
        ds.close()
