#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy as np
import pandas as pd
import netCDF4 as nc
import os
import itertools

from tqdm import tqdm
from LatLon23 import LatLon, Latitude, Longitude
from sklearn.cluster import KMeans

import settings
from data_preprocessing import generate_data_matrix
from utilities.import_format import import_nc, import_tif
from utilities.cut_and_interpolate import cut_and_interpolate
from utilities.strings_for_ncfile import char_to_string
from utilities.initialise_log import save_log
from utilities.import_raw_dataset import import_raw_dataset
from utilities.handle_categorical_values import handle_categorical_values


class create_training_data:

    """
        This class generates the training dataset
        for the Random Forest Classifier.

        Input:
            from_scratch: boolean, True if training dataset should be
                          generated from scratch, otherwise False
            delete: True if dataset/feature should be deleted from csv-file
                    False if dataset should be added to existing csv-file
                    (careful: from_scratch needs to be False!)
            data_to_handle: list of features that should be added/deleted
                            datasets need to be listed in list_of_raw_datasets
            preprocessing: either 'cluster' (boolean, True if training locations are to be clustered),
                           'interpolation' (boolean, True if datasets are supposed to be
                            interpolated before extracting information) or 'no_interpolation' 
                           (boolean, True if no interpolation shall be applied)
            log: logger to add information to shire_run.log

        Output:
            csv file
    """

    def __init__(self,
                 from_scratch=True,
                 delete=True,
                 data_to_handle=None,
                 preprocessing=None,
                 log=None):

        self.from_scratch = from_scratch
        self.delete = delete
        self.data_to_handle = data_to_handle
        self.num_clusters = 15
        if preprocessing == 'cluster':
            self.cluster = True
            self.interpolation = True
        elif preprocessing == 'interpolation':
            self.cluster = False
            self.interpolation = True
        elif preprocessing == 'no_interpolation':
            self.cluster = False
            self.interpolation = False
        self.df_train = 'not_set_yet'
        self.logger = log

        self.import_data_summary()
        self.data_to_handle = self.keys_to_include
        
        if self.from_scratch:
            if self.cluster is True:    
                self.main_cluster()
            elif self.interpolation is True:
                self.main()
            elif self.interpolation is False and self.cluster is False:
                self.main_no_interpolation()
        elif not self.from_scratch and self.delete:
            # Import existing training dataset
            self.import_input_training_dataset()
            self.delete_feature()   # Delete feature from training dataset
            self.save_training_data()   # Save training dataset as csv

        elif not self.from_scratch and not self.delete:
            # Import existing training dataset
            self.import_input_training_dataset()
            self.add_feature()  # Add feature to the training dataset
            self.save_training_data()  # Save training dataset as csv
                
    def import_data_summary(self):
        
        """
            Import the information on the geospatial datasets and the 
            keys to include in the training dataset
        """
        
        self.data_properties = pd.read_csv(settings.data_summary_path)
        self.keys_to_include = pd.read_csv(settings.key_to_include_path)['keys_to_include'].tolist()
        self.logger.info('Data summary in keys to include have been imported')
        
    def handle_categorical_features(self, var=None):
        
        """
            Function is called which performs one-hot or ordinal encoding
        """
        
        basic = ['ID', 'xcoord', 'ycoord', 'label']
        self.df_train = handle_categorical_values(self.df_train,
                                                  self.data_properties,
                                                  settings.ohe,
                                                  basic,
                                                  var)

    def delete_feature(self):

        """
            Features in data_to_handle are deleted from the training dataset
        """
        self.logger.info(
                        'Feature(s) will be deleted')
        to_drop = []
        for feat in self.data_to_handle:
            for col in self.df_train.columns.tolist():
                if feat in col:
                    to_drop.append(col)
            
        self.df_train.drop(columns=to_drop, inplace=True)
        
        self.logger.info(
                        'Feature has been deleted')
        self.logger.info(
                        'Training dataset contains the following features: '
                        + str(self.df_train.columns))
        print('Training dataset contains following features: '
              + str(self.df_train.columns))

    def import_cube(self):

        """
            Import data cube created in data_preprocessing.py
        """
        self.logger.info(
                        'Import the cube with interpolated geospatial data')
        
        if not os.path.exists(self.s.outfile):
            self.logger.error(
                        'Cube could not be found!')
        
        self.ds = nc.Dataset(self.s.outfile)
        self.x = self.ds['Longitude'][:].data
        self.y = self.ds['Latitude'][:].data
        self.cube = self.ds['Result'][:, :, :].data

        features = self.ds['features'][:].data

        self.features = char_to_string(features)
        self.logger.info(
                        'Cube contains the following features: ' 
                        + str(self.features))
        print('Cube contains following features: ' + str(self.features))

    def import_landslide_database(self):

        """
            Import training dataset which needs to be provided as csv file
            The landslide instances are supplemented with absence locations.
            Missing values and Nan values are removed.
        """

        self.df_train = pd.read_csv(settings.path_landslide_database)

        self.logger.info('Landslide inventory imported')
        for column in list(self.df_train.columns):
            if column not in [settings.ID,
                              settings.landslide_database_x,
                              settings.landslide_database_y]:
                self.df_train.drop(inplace=True, labels=column, axis=1)
        self.check_ls_inventory()
        self.add_nonls_locations()
        self.length_before_cleaning = len(self.df_train)

        self.df_train = pd.concat([self.df_train, self.df_absence],
                                  axis=0,
                                  ignore_index=True)
        self.logger.info('Absence locations added')
        self.df_train = self.df_train.rename(
            columns={settings.landslide_database_x: 'xcoord',
                     settings.landslide_database_y: 'ycoord',
                     settings.ID: 'ID'})
        self.df_train['label'] = self.label_training_data()
        self.logger.info('Label added')

    def add_nonls_locations(self):

        """
            Supplement presence data with absence data. It needs to be
            pre-generated.
        """

        ds = nc.Dataset(settings.path_nonls_locations)

        x = ds[settings.nonls_database_x][:].data
        y = ds[settings.nonls_database_y][:].data
        
        self.df_absence = pd.DataFrame(index=range(len(x)),
                                       columns=list(self.df_train.columns))

        self.df_absence[settings.ID] = ['nonls_event_' + str(i)
                                        for i in range(len(x))]
        self.df_absence[settings.landslide_database_x] = list(x)
        self.df_absence[settings.landslide_database_y] = list(y)

    def label_training_data(self):

        """
            Provide labels to the training data
        """

        label = [1 for i in range(np.shape(self.df_train)[0])]
        label[
            self.length_before_cleaning:len(label)] = \
            [0 for i in range(len(label)-self.length_before_cleaning)]

        return label

    def import_input_training_dataset(self):

        """
            Existing training dataset is imported.
        """
        
        self.logger.info('Existing training dataset will be imported')
        # Check if training dataset exists and import it
        if not os.path.exists(settings.path_train):
            print('Training dataset does not exist yet.\
                  Please generate from scratch.')
            self.logger.error('Training dataset could not be found!')
        else:
            self.df_train = pd.read_csv(settings.path_train)
            self.features = list(self.df_train.columns)
            print('Training dataset consists the following features: '
                  + str(self.features))
            self.logger.info('Training dataset contains the following features: '
                             + str(self.features))
            self.length_before_cleaning = len(self.df_train[self.df_train['label'] == 1])
            
    def add_feature(self):

        """
            Add feature to the training dataset
        """

        self.logger.info('Feature adding started')

        # Features contained in the training dataset
        cols = list(self.df_train.columns)
        # List of labels for basic information
        basic = ['Ereignis-Nr', 'xcoord', 'ycoord', 'label']

        if self.interpolation:
            # Call data_preprocessing.py
            self.s = generate_data_matrix(
                from_scratch=False,
                delete=False,
                dataset='training',
                keys_already_included=[x for x in cols if x not in basic])

            # Import generated cube of cut and interpolated datasets
            self.import_cube()

            for count, key in enumerate(self.data_to_handle):
                # Delete feature if it already exists in training dataset
                if key in self.df_train.columns:
                    self.logger.warning('Feature already exists in Dataset.\
                                        Existing feature is deleted')
                    self.df_train = self.df_train.drop(key, axis=1)

                self.logger.info('Adding ' + key + '...')
                # Create empty DataFrame
                if count == 0:
                    self.df_features = pd.DataFrame(
                        index=range(len(self.df_train)),
                        columns=self.data_to_handle)

                for index, row in self.df_train.iterrows():
                    x_ind = int(
                        (np.abs(
                            self.x
                            - row[settings.landslide_database_x])).argmin())
                    y_ind = int(
                        (np.abs(
                            self.y
                            - row[settings.landslide_database_y])).argmin())

                    self.df_features.at[index, key] = self.cube[
                        y_ind,
                        x_ind,
                        self.features.index(key)]
            # Combine old training dataset with additional features
            self.df_train = pd.concat([self.df_train, self.df_features],
                                      axis=1)
            # Adapt column order
            self.logger.info('Feature successfully added')

            self.clean_df()
            self.handle_categorical_features(var=self.data_to_handle)
            self.logger.info('One-hot encoding completed')
            
        elif self.cluster:
            for key in self.data_to_handle:
                if key in self.df_train.columns:
                    self.logger.info(
                        'Feature already exists in dataset.\
                            Existing feature is deleted')
                    self.df_train = self.df_train.drop(key, axis=1)

            self.main_cluster()
            self.logger.info('Feature successfully added')
        elif not self.cluster and not self.interpolation:
            self.main_no_interpolation()
            self.logger.info('Feature successfully added')

        self.logger.info('Training dataset contains following features:')
        self.logger.info(str(self.df_train.columns))

    def extract_gridded_info(self):

        """
            Extraction of the information of the geospatial datasets at all
            elements of the training dataset.

            If training data is located within prediction area and if area is
            small enough no further interpolation is necessary.
        """

        self.df_features = pd.DataFrame(index=range(len(self.df_train)),
                                        columns=self.features)

        # Iterate over all instances of the training dataset
        # and extract geospatial information
        for index, row in tqdm(self.df_train.iterrows()):
            # Indices of training data elements are determined
            x_ind = int((np.abs(self.x - row['xcoord'])).argmin())
            y_ind = int((np.abs(self.y - row['ycoord'])).argmin())

            tmp = list(self.cube[y_ind, x_ind, :])
            self.df_features.loc[index] = tmp

    def check_ls_inventory(self):

        """
            Rows are removed with missing values or nan
        """
        
        self.logger.info('Landslide inventory is check for missing and nan values')
        rows_with_missing_values = self.df_train[self.df_train.isnull().any(axis=1)].index.tolist()
        
        if len(set(rows_with_missing_values)) > 0:
            self.df_train.drop(index=set(rows_with_missing_values), inplace=True)
            self.logger.info(str(len(set(rows_with_missing_values)))
                             + ' rows are removed due to missing or nan values')

    def clean_df(self):

        """
            Rows are removed from the dataset where no_value given in settings
            occurs.
        """
        
        self.logger.info('Clean the dataframe from no data instances')
        count = 0
        ind = []
        for col in self.df_train:
            ind.append([i for i, x in enumerate(list(self.df_train[col]))
                        if x == settings.no_value])
            count = count + 1

        ind = list(itertools.chain.from_iterable(ind))
        ind = list(set(ind))

        self.logger.info(str(len(ind)) + ' rows will be removed')
        print(str(len(ind)) + ' rows will be removed due to invalid data')

        self.df_train.drop(index=ind, inplace=True)
        self.df_train.reset_index(drop=True, inplace=True)

        if not self.from_scratch and len(ind) > 0:
            print('Careful! Ratio might be obscured!')
            self.logger.warning('Careful! Ratio might be obscured!')

    def ensure_same_ratio(self):

        """
            Ensure that the desired ratio of presence to absence data is
            kept even after removing instances with no data values
        """

        self.logger.info('Ensure that the ratio of presence to absence is kept')
        len_pres = np.shape(self.df_train[self.df_train['label'] == 1])[0]
        len_abs = np.shape(self.df_train[self.df_train['label'] == 0])[0]

        if settings.num_nonls == self.length_before_cleaning:
            if len_abs > len_pres:
                print('Number of absence locations is reduced')
                self.logger.info('Number of absence locations is reduced')
                df_abs = self.df_train[self.df_train['label'] == 0]

                df_abs = df_abs.iloc[:len_abs-(len_abs-len_pres)]

                self.df_train = pd.concat(
                    [self.df_train[self.df_train['label'] == 1], df_abs],
                    axis=0)
            elif len_abs < len_pres:
                print(
                    'Undefined error in the number\
                        of absence and presence data')
                self.logger.error('Undefined error in the number of absence and presence instances')
        else:
            if len_abs > settings.num_nonls:
                df_abs = self.df_train[self.df_train['label'] == 0]

                df_abs = df_abs.iloc[:len_abs
                                     - (len_abs
                                        - settings.num_nonls)]

                self.df_train = pd.concat(
                    [self.df_train[self.df_train['label'] == 1], df_abs],
                    axis=0)
                self.logger.info('Surplus absence locations have been removed')

    def cluster_landslide_locations(self, re=False, num=None):

        """
            Cluster the landslide locations. If clusters are too large this
            functions reclusters these clusters into smaller ones.

            Input:
                re: boolean, True if reclustering
                num: list, numbers of clusters to be reclustered

            Output:
                re_cluster_name: list, clusters for every entry
        """

        if re:
            re_cluster_name = []
            count, len_org = 0, len(self.bb)

            # Number of clusters to split too large cluster into
            num_cluster = 4
            for i in num:

                df_sub = self.df_train[self.df_train.cluster == i].copy()
                # Clustering
                kmeans = KMeans(init="random",
                                n_clusters=num_cluster,
                                n_init=10,
                                max_iter=300,
                                random_state=42)
                kmeans.fit(np.column_stack((list(df_sub['xcoord']),
                                            list(df_sub['ycoord']))))
                tmp = kmeans.labels_[:]

                # Rename clusters to fit into naming convention
                for c, j in enumerate(tmp):
                    if j == 0:
                        tmp[c] = i
                    else:
                        tmp[c] = len_org + count*(num_cluster-1) + (j-1)

                df_sub['cluster'] = tmp
                self.df_train = pd.concat(
                    [self.df_train[self.df_train.cluster != i], df_sub],
                    axis=0)
                # Store cluster numbers to be returned
                re_cluster_name.append(set(tmp))
                count = count + 1

            return re_cluster_name

        else:
            print('Start clustering...')
            self.logger.info('Start clustering...')
            # Clustering
            kmeans = KMeans(init="random",
                            n_clusters=self.num_clusters,
                            n_init=10,
                            max_iter=300,
                            random_state=42)
            kmeans.fit(np.column_stack((list(self.df_train['xcoord']),
                                        list(self.df_train['ycoord']))))
            self.df_train['cluster'] = kmeans.labels_[:]

    def determine_bb_for_clustering(self, re=False, re_num=None):

        """
            Determine bounding box for the individual clusters.

            Input:
                re: boolean, True if reclustering
                num: list, numbers of clusters to be reclustered
        """

        print('Determine bounding boxes...')
        self.logger.info('Determine bounding box(es)...')

        if self.cluster:
            # Initial clustering
            if not re:
                self.bb = []
                if self.from_scratch:
                    to_enter = range(self.num_clusters)
                else:
                    # When adding a feature
                    if 'cluster' in self.df_train:
                        to_enter = list(set(self.df_train.cluster.to_list()))
                    else:
                        to_enter = range(self.num_clusters)
                for num in to_enter:
                    df_tmp = self.df_train[
                        self.df_train['cluster'] == num].copy()

                    df_tmp = df_tmp.reset_index(drop=True)
                    max_x = df_tmp['xcoord'].loc[df_tmp['xcoord'].idxmax()]
                    min_y = df_tmp['ycoord'].loc[df_tmp['ycoord'].idxmin()]
                    max_y = df_tmp['ycoord'].loc[df_tmp['ycoord'].idxmax()]
                    min_x = df_tmp['xcoord'].loc[df_tmp['xcoord'].idxmin()]

                    self.bb.append([max_y, min_y, min_x, max_x])
            # Reclustering
            else:
                bb_new = [0 for i in range(
                    len(set(self.df_train.cluster.to_list())))]
                bb_new[:len(self.bb)] = self.bb

                for num in re_num:

                    df_tmp = self.df_train[
                        self.df_train['cluster'] == num].copy()
                    df_tmp = df_tmp.reset_index(drop=True)
                    max_x = df_tmp['xcoord'].loc[df_tmp['xcoord'].idxmax()]
                    min_y = df_tmp['ycoord'].loc[df_tmp['ycoord'].idxmin()]
                    max_y = df_tmp['ycoord'].loc[df_tmp['ycoord'].idxmax()]
                    min_x = df_tmp['xcoord'].loc[df_tmp['xcoord'].idxmin()]

                    # Make sure that the order is preserved
                    # to match bounding boxes properly
                    if num >= self.num_clusters:
                        bb_new[num] = [max_y, min_y, min_x, max_x]
                    else:
                        bb_new[num] = [max_y, min_y, min_x, max_x]

                self.bb = bb_new.copy()

        else:
            max_x = self.df_train['xcoord'].loc[
                self.df_train['xcoord'].idxmax()]
            min_y = self.df_train['ycoord'].loc[
                self.df_train['ycoord'].idxmin()]
            max_y = self.df_train['ycoord'].loc[
                self.df_train['ycoord'].idxmax()]
            min_x = self.df_train['xcoord'].loc[
                self.df_train['xcoord'].idxmin()]

            self.bb = [max_y, min_y, min_x, max_x]
            print(self.bb)
            self.logger.info('Bounding box of training samples: ' 
                             + str(max_y) + ', ' + str(min_y) + ', '
                             + str(min_x) + ', '+ str(max_x))

    def determine_if_reclustering(self):

        """
            Determine if the extent of one or several
            clusters are too large for local interpolation

            Output:
                num_bb: list, names of clusters that need reclustering
        """

        self.reclustering = False
        num_bb = []

        # Check extend of individual clusters
        for count, bb in enumerate(self.bb):
            point1_x = LatLon(Latitude(bb[0]), Longitude(bb[2]))
            point2_x = LatLon(Latitude(bb[0]), Longitude(bb[3]))
            distance_x = point1_x.distance(point2_x)*1000

            point1_y = LatLon(Latitude(bb[0]), Longitude(bb[2]))
            point2_y = LatLon(Latitude(bb[-1]), Longitude(bb[2]))
            distance_y = point1_y.distance(point2_y)*1000

            num_px_x = int(np.round((distance_x/settings.resolution)))
            num_px_y = int(np.round((distance_y/settings.resolution)))

            if num_px_x or num_px_y > 10000:
                num_bb.append(count)
                self.reclustering = True

        return num_bb

    def main_cluster(self):

        """
            Main function to generate training dataset if training
            locations shall be clustered.
        """

        def extract_data_from_dataset_subsets(num):

            """
                Extract information from the interpolated geospatial dataset.

                Input:
                    num: int, number of cluster

                Output:
                    df_clust: dataframe, subset of the training dataset
                    supplemented with the information from the training dataset
            """

            df_clust = self.df_train[self.df_train.cluster == num].copy()
            #print(len(df_clust[df_clust['label'] == 0]), len(df_clust[df_clust['label'] == 1]))

            # Access interpolated subset of the dataset
            arr = self.ds['Result' + str(num)][:, :].data

            # Access interpolation x and y vector
            yvector = self.ds['Latitude' + str(num)][:].data
            xvector = self.ds['Longitude' + str(num)][:].data

            # Extract information at training location
            feature = []
            for index, row in df_clust.iterrows():
                x_indx = int((np.abs(np.array(xvector)
                                     - row['xcoord'])).argmin())
                y_indx = int((np.abs(np.array(yvector)
                                     - row['ycoord'])).argmin())

                feature.append(arr[y_indx, x_indx])

            df_clust[dataset] = feature

            return df_clust

        self.logger.info('Approach: clustering')
        # If previous run failed dataset might already exist causing
        # current run to crash again therefore it is deleted if it exists
        if os.path.isfile('tmp.nc'):
            os.remove('tmp.nc')

        if self.from_scratch:
            self.logger.info('Training dataset generated from scratch')
            if not isinstance(self.df_train, pd.DataFrame):
                self.import_landslide_database()  # Import landslide database
            self.cluster_landslide_locations()

        if not self.from_scratch and 'cluster' not in self.df_train:
            # Cluster the landslide locations
            self.logger.info('Existing training dataset has not been generated using clustering')
            self.cluster_landslide_locations()
        # Determine the bounding boxes for the individual clusters
        self.determine_bb_for_clustering()

        if self.from_scratch:
            # Determine if bounding boxes are too large for local interpolation
            num_bb = self.determine_if_reclustering()

            if self.reclustering:
                print('Reclustering necessary...')
                self.logger.info('Reclustering necessary...')
                re_cluster_name = self.cluster_landslide_locations(True,
                                                                   num_bb)
                re_cluster_name = [item for sublist in re_cluster_name
                                   for item in sublist]
                # Determine bounding boxes for new clusters
                self.determine_bb_for_clustering(True, re_cluster_name)

        dataset_counter = 0
        # Iterate over the dataset to inclue in the training dataset
        for dataset in tqdm(self.data_to_handle):
            self.logger.info('Currently handling: ' + dataset)
            index = list(self.data_properties['keys']).index(dataset)
            # Call cut_and_interpolate class to cut
            # and interpolate the current dataset

            s = cut_and_interpolate(
                key=dataset,
                path=list(self.data_properties['path'])[index],
                no_data_value=list(self.data_properties['no_value'])[index],
                categorical=bool(list(self.data_properties['categorical'])[index]),
                several=False,
                several_same=False,
                first=False,
                bb=self.bb,
                cluster=self.cluster,
                path_properties=settings.path_ml + 'properties.pkl')
            # Open the netcdf file which contains the interpolated subsets of
            # the dataset with the extent of the bounding boxes of the clusters
            self.logger.info('Interpolation completed')
            self.ds = nc.Dataset('tmp.nc')

            # Determine if one or more bounding boxes is
            # outside of the extend of the dataset
            if not s.cuttable:
                print('Error! Bounding box larger than dataset!\
                      Please adapt bounding_box!')
                self.logger.error('Error! Bounding box larger than dataset!')
                break

            df = []
            # Iterate over the clusters extract information from the dataset
            for num in range(len(self.bb)):
                df.append(extract_data_from_dataset_subsets(num))

            # Concatenate all subsets of the training dataset
            df = np.concatenate(df, axis=0)
            self.logger.info('Extraction of values from dataset completed')
            # For first dataset set up final training dataset,
            # for later datasets append information
            if dataset_counter == 0:
                df_ges = pd.DataFrame(df, columns=list(self.df_train.columns)
                                      + [dataset])
            else:
                df_ges[dataset] = df[:, -1]
                

            dataset_counter = dataset_counter + 1
            self.ds.close()

        # Remove temporary netcdf file that contains
        # interpolated subsets of the dataset
        os.remove('tmp.nc')

        self.df_train = df_ges.copy()
        self.df_train.reset_index(inplace=True, drop=True)

        self.clean_df()  # Clean the DataFrame from no value rows
        if not self.from_scratch and not self.delete:
            self.handle_categorical_features(var=self.data_to_handle)
        else:
            self.handle_categorical_features()
        if self.from_scratch:
            # Ensure that the 1:1 ratio of the training dataset is perserved
            self.ensure_same_ratio()

        self.save_training_data()  # Save the training dataset as csv file

    def main(self):
        
        self.logger.info('Approach: interpolation')
        if not isinstance(self.df_train, pd.DataFrame):
            self.import_landslide_database()  # Import landslide database
        
        print('Warning! Depending on the size of the area of interest and\
              the set resolution this might be computationally expensive.\
                  Consider clustering.')
        self.logger.warning('Warning! Depending on the size of the AOI and\
                            the set resolution this might be computationally\
                            expensive. Consider clustering')
        self.logger.info('Cube of interpolated datasets will be generated')
        self.determine_bb_for_clustering(re=False, re_num=None)
        self.s = generate_data_matrix(from_scratch=True,
                                      delete=False,
                                      dataset='training',
                                      bb=self.bb)
        self.logger.info('Cube of interpolated datasets successfully generated')
        self.import_cube()
        # Extract geospatial information at ls and non-ls locations
        self.extract_gridded_info()
        self.logger.info('Gridded information has been extraced')

        # Concat final training dataset
        self.df_train = pd.concat([self.df_train, self.df_features], axis=1)

        self.clean_df()  # Clean the DataFrame from no value rows
        if not self.from_scratch and not self.delete:
            self.handle_categorical_features(var=self.data_to_handle)
        elif self.from_scratch:
            print('ohe')
            self.handle_categorical_features()
        self.ensure_same_ratio()
        self.save_training_data()  # Save the training dataset as csv file

    def raw_dataset(self, path, no_data):

        if no_data != 'None':
            no_data = no_data.split(',')
            no_data = [float(val) for val in no_data]

        data, x, y = import_raw_dataset(path, no_data, settings.no_value)
        if isinstance(data, np.ndarray):
            self.logger.info('Raw dataset imported')
        else:
            print(
                'Not the right data format! Please provide tif or nc file')
            self.logger.info('Wrong file format!')

        return data, x, y

    def main_no_interpolation(self):

        print('no interpolation...')
        self.logger.info('Approach: no interpolation')

        def extract_gridded_info(row):

            x_indx = int((np.abs(np.array(x) - row['xcoord'])).argmin())
            y_indx = int((np.abs(np.array(y) - row['ycoord'])).argmin())

            return data[y_indx, x_indx]

        if self.from_scratch:
            self.import_landslide_database()
        else:
            if not isinstance(self.df_train, pd.DataFrame):
                self.import_landslide_database()  # Import landslide database

        for dataset in tqdm(self.data_to_handle):
            self.logger.info('Currently handling: ' + dataset)
            index = list(self.data_properties['keys']).index(dataset)
            data, x, y = self.raw_dataset(self.data_properties['path'][index],
                                                 self.data_properties['no_value'][index])
            feat = []
            for index, row in self.df_train.iterrows():
                feat.append(extract_gridded_info(row))

            self.df_train[dataset] = feat
            self.logger.info('Gridded information has been extracted')
            
        self.clean_df()  # Clean the DataFrame from no value rows
        if not self.from_scratch and not self.delete:
            self.handle_categorical_features(var=self.data_to_handle)
        else:
            self.handle_categorical_features()
            self.ensure_same_ratio()
        self.save_training_data()

    def save_training_data(self):

        """
            Save dataframe as csv. If necessary folder is created.
        """
        
        self.df_train = self.df_train.rename(columns={settings.ID: 'ID', settings.landslide_database_x: 'xcoord', settings.landslide_database_y: 'ycoord'})
        
        if 'training.csv' not in settings.path_train:
            outfile = settings.path_train + 'training.csv'
        else:
            outfile = settings.path_train

        isExist = os.path.exists(outfile.split('/')[-1])
        if not isExist:
            os.makedirs(outfile.split('/')[-1])

        # If outfile exists already, delete
        if os.path.exists(outfile):
            os.remove(outfile)

        # Save dataframe as csv
        self.df_train.to_csv(outfile, sep=',', index=False)
        self.logger.info('Training dataset has been saved to '
                         + outfile)
