#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy as np
from tqdm import tqdm
import os
import netCDF4 as nc
import settings
import pandas as pd

import settings
from utilities.ncfile_generation import generate_3dncfile
from utilities.cut_and_interpolate import cut_and_interpolate
from utilities.strings_for_ncfile import features_to_char, char_to_string

class generate_data_matrix:

    """
        This class generates a nc-file containing all datasets,
        a list of all contained features and their respective longitude and
        latitude vectors. Provided are interpolated excerpts of the datasets
        determined by the provided bounding box.

        Input:
            from_scratch: boolean, True if nc-file should be generated from
                          scratch, otherwise false
            delete: True if dataset/feature should be deleted from nc-file
                    False if dataset should be added to existing nc-file
                    (careful: from_scratch needs to be False!)
            bb: list, bounding box in the format
            [<y_max>, <y_min>, <x_min>, <x_max>]
            data_to_handle: list of features that should be added/deleted
                            datasets need to be listed in list_of_raw_datasets
            keys_already_included: list of already included features in
                                    the training/prediction dataset
                                  (from_scratch=False, delete=False)

        Output:
            netCDF4 file
    """

    def __init__(self,
                 from_scratch=True,
                 delete=False,
                 dataset='undefined',
                 bb=None,
                 data_to_handle=None,
                 keys_already_included=None):

        self.from_scratch = from_scratch
        self.dataset = dataset
        self.bb = bb
        self.keys_already_included = keys_already_included
        
        self.import_data_summary()

        if not from_scratch:
            self.data_to_handle = self.keys_to_include
            self.delete = delete
            self.import_cube()

            if delete:
                # Delete dataset from cube
                self.delete_dataset()
            else:
                # Add dataset to cube
                self.add_dataset()
        else:
            # Generate cube from scratch
            # Parameter stored in list_of_raw_datasets
            self.data_to_handle = self.keys_to_include
            self.main()
            
    def import_data_summary(self):
        
        """
            Import the csv files containing information on the geospatial datasets
            and the keys to include
        """
        
        self.data_properties = pd.read_csv(settings.data_summary_path)
        self.keys_to_include = pd.read_csv(settings.key_to_include_path)['keys_to_include'].tolist()

    def find_dataset(self):

        """
            Find the index of the features to handle in the list of features
            contained in the nc-file.
            Return:
                idx: list of indices
        """

        return self.features.index(self.data_to_handle)
    
    def add_dataset(self):
        
        # Number of overlapping features between datasets in the cube
        # and the datasets to add/delete
        print('features')
        print(self.features)
        
        if self.dataset == 'prediction':
            for_prediction = True
        else:
            for_prediction = False

        # Define new cube in the size of existing cube of cut and interpolated
        # datasets with the depth equaling the number of existing datasets plus
        # the ones to add
        ges = list(self.features) + [x for x in self.data_to_handle if x not in self.features]
        cube = np.zeros((np.shape(self.cube)[0],
                         np.shape(self.cube)[1],
                         len(ges)))
        
        for feat in self.features:
            cube[:, :, ges.index(feat)] = self.cube[:, :, self.features.index(feat)]
        
        for key in self.data_to_handle:
            s = cut_and_interpolate(
                key=key,
                path=list(self.data_properties['path'])[list(self.data_properties['keys']).index(key)],
                no_data_value=list(self.data_properties['no_value'])[list(self.data_properties['keys']).index(key)],
                categorical=list(self.data_properties['categorical'])[list(self.data_properties['keys']).index(key)],
                several=True,
                several_same=False,
                first=False,
                #bb=self.bb,
                for_prediction=for_prediction,
                path_properties=self.folder
                + '/data_combined_'
                + self.dataset
                + '_'
                + str(settings.resolution) + '.pkl') 
            array, _, _, cuttable = s.array, s.x, s.y, s.cuttable
            
            if not cuttable:
                print('Error! Bounding box larger than dataset!\
                          Please adapt bounding_box!')
                break
            else:
                # Store it at respective position in cube
                # Add cut and interpolated dataset to cube
                cube[:, :, ges.index(key)] = array
                
        # Save the updated cube to nc file
        self.determine_outfile()
        self.char_features = features_to_char(ges)
        generate_3dncfile(self.outfile,
                          self.x,
                          self.y,
                          cube,
                          len(ges),
                          self.char_features,
                          crs='wgs84',
                          data_unit=None,
                          missing_value=settings.no_value)

    def delete_dataset(self):

        """
            Delte datasets from data_to_handle
            from nc-file and save new nc-file
        """

        # Determine indices of the datasets that shall be removed
        idx = []
        for data in self.data_to_handle:
            idx.append(self.find_dataset)

        # Define new cube in the size of existing
        # cube of cut and interpolated datasets
        cube = np.zeros((np.shape(self.cube)[0],
                         np.shape(self.cube)[1],
                         np.shape(self.cube)[2]-len(self.data_to_handle)))
        count = 0

        # Go through the datasets and transfer all
        # datasets except for them to be removed
        for i in range(np.shape(self.cube)[2]):
            if self.features[i] not in self.data_to_handle:
                cube[:, :, count] = self.cube[:, :, i]
                count = count + 1

        # Update the feature list
        for data in self.data_to_handle:
            self.features.remove(data)
        print('Dataset now contains the following features: '
              + str(self.features))

        # Save new data cube
        self.determine_outfile()
        self.char_features = features_to_char(self.from_scratch, self.features)
        generate_3dncfile(self.outfile,
                          self.x,
                          self.y,
                          cube,
                          len(self.features),
                          self.char_features,
                          crs='wgs84',
                          data_unit=None,
                          missing_value=settings.no_value)

    def import_cube(self):

        """
            Existing nc-file is imported for adding/deleting another feature.
        """

        self.determine_outfile()  # Determine where cube is stored

        # Import cube
        self.ds = nc.Dataset(self.outfile)
        self.cube = self.ds['Result'][:, :, :].data
        self.x = self.ds['Longitude'][:].data
        self.y = self.ds['Latitude'][:].data
        self.features = self.ds['features'][:].data

        self.features = char_to_string(self.features)
        print('Dataset contains the following features: ' + str(self.features))

    def determine_outfile(self):

        """
            Determine whether folder to store the nc-file already exists.
            If not, it is created. Outfile path is determined.
        """

        # Cube is stored in the same folder
        # as the final training/prediction dataset
        if self.dataset == 'training':
            self.folder = settings.path_train.rsplit('/', 1)[0]
            self.outfile = self.folder + '/data_combined_training_' + str(settings.resolution) + '.nc'
        elif self.dataset == 'prediction':
            self.folder = settings.path_pred.rsplit('/', 1)[0]
            self.outfile = self.folder + '/data_combined_prediction_' + str(settings.resolution) + '.nc'

        # Create folder if it doesn't yet exist
        isExist = os.path.exists(self.folder)
        if not isExist:
            os.makedirs(self.folder)

    def main(self):

        """
            Routine to pre-process the datasets from scratch
        """

        # Go through all datasets that shall be pre-processed
        for i in tqdm(range(len(self.data_to_handle))):
            j = list(self.data_properties['keys']).index(self.data_to_handle[i])

            if i == 0:
                if self.dataset == 'prediction':
                    # Cut and interpolate dataset to desired resolution.
                    # Check script for information on input parameters.
                    s = cut_and_interpolate(
                        key=self.data_to_handle[i],
                        path=self.data_properties['path'].tolist()[j],
                        no_data_value=self.data_properties['no_value'].tolist()[j],
                        categorical=bool(self.data_properties['categorical'].tolist()[j]),
                        several=True,
                        several_same=False,
                        first=True,
                        bb=self.bb,
                        for_prediction=True,
                        path_properties=settings.path_pred.rsplit('/', 1)[0] + '/data_combined_prediction_' + str(settings.resolution) + '.pkl')
                else:
                    # Cut and interpolate dataset to desired resolution.
                    # Check script for information on input parameters.
                    s = cut_and_interpolate(
                        key=self.data_to_handle[i],
                        path=self.data_properties['path'].tolist()[j],
                        no_data_value=self.data_properties['no_value'].tolist()[j],
                        categorical=bool(self.data_properties['categorical'].tolist()[j]),
                        several=True,
                        several_same=False,
                        first=True,
                        bb=self.bb,
                        path_properties=settings.path_train.rsplit('/', 1)[0] + '/data_combined_training_' + str(settings.resolution) + '.pkl')
                array = s.array
                self.x = s.x
                self.y = s.y
                cuttable = s.cuttable

                if not cuttable:
                    print('Error! Bounding box larger than dataset!\
                          Please adapt bounding_box!')
                    break

                # Store cut and interpolated dataset in array
                cube = np.zeros((np.shape(array)[0],
                                 np.shape(array)[1],
                                 len(self.data_to_handle)))
                cube[:, :, 0] = array
            else:
                if self.dataset == 'prediction':
                    s = cut_and_interpolate(
                        key=self.data_to_handle[i],
                        path=self.data_properties['path'].tolist()[j],
                        no_data_value=self.data_properties['no_value'].tolist()[j],
                        categorical=bool(self.data_properties['categorical'].tolist()[j]),
                        several=True,
                        several_same=False,
                        first=False,
                        bb=self.bb,
                        for_prediction=True,
                        path_properties=settings.path_pred.rsplit('/', 1)[0]
                        + '/data_combined_prediction_'
                        + str(settings.resolution)
                        + '.pkl')
                else:
                    # Cut and interpolate dataset to desired resolution.
                    # Check script for information on input parameters.
                    s = cut_and_interpolate(
                        key=self.data_to_handle[i],
                        path=self.data_properties['path'].tolist()[j],
                        no_data_value=self.data_properties['no_value'].tolist()[j],
                        categorical=bool(self.data_properties['categorical'].tolist()[j]),
                        several=True,
                        several_same=False,
                        first=False,
                        bb=self.bb,
                        path_properties=settings.path_train.rsplit('/', 1)[0]
                        + '/data_combined_training_'
                        + str(settings.resolution)
                        + '.pkl')
                array, cuttable = s.array, s.cuttable

                if not cuttable:
                    print('Error! Bounding box larger than dataset!\
                          Please adapt bounding_box!')
                    break
                # Store cut and interpolated dataset in array
                cube[:, :, i] = array

            # Store the array in a nc-file and meta data in pickle file
            if cuttable:
                self.determine_outfile()
                self.char_features = features_to_char(self.data_to_handle)
                generate_3dncfile(self.outfile,
                                  self.x,
                                  self.y,
                                  cube,
                                  len(self.data_to_handle),
                                  self.char_features,
                                  crs='wgs84',
                                  data_unit=None,
                                  missing_value=settings.no_value)
